"""
This module implements utilities to convert between RA, Dec and indexes
on the Hierarchical Triangular Mesh (HTM), a system of tiling the unit sphere
with nested triangles.  The HTM is described in these references

Kunszt P., Szalay A., Thakar A. (2006) in "Mining The Sky",
Banday A, Zaroubi S, Bartelmann M. eds.
ESO Astrophysics Symposia
https://www.researchgate.net/publication/226072008_The_Hierarchical_Triangular_Mesh

Szalay A. et al. (2007)
"Indexing the Sphere with the Hierarchical Triangular Mesh"
arXiv:cs/0701164
"""

__all__ = (
    "Trixel",
    "HalfSpace",
    "find_htmid",
    "trixel_from_htmid",
    "basic_trixels",
    "half_space_from_ra_dec",
    "level_from_htmid",
    "get_all_trixels",
    "half_space_from_points",
    "intersect_half_spaces",
)

import numbers

import numpy as np

from rubin_scheduler.utils import cartesian_from_spherical, spherical_from_cartesian


class Trixel:
    """
    A trixel is a single triangle in the Hierarchical Triangular Mesh (HTM)
    tiling scheme.  It is defined by its three corners on the unit sphere.

    Instantiating this class directly is a bad idea. __init__() does nothing
    to ensure that the parameters you give it are self-consistent.  Instead,
    use the trixel_from_htmid() or get_all_trixels() methods in this module
    to instantiate trixels.
    """

    def __init__(self, present_htmid, present_corners):
        """
        Initialize the current Trixel

        Parameters
        ----------
        present_htmid is the htmid of this Trixel

        present_corners is a numpy array.  Each row
        contains the Cartesian coordinates of one of
        this Trixel's corners.

        WARNING
        -------
        No effort is made to ensure that the parameters
        passed in are self consistent.  You should probably
        not being calling __init__() directly.  Use the
        trixel_from_htmid() or get_all_trixels() methods to
        instantiate trixels.
        """
        self._corners = present_corners
        self._htmid = present_htmid
        self._level = (len("{0:b}".format(self._htmid)) / 2) - 1
        self._cross01 = None
        self._cross12 = None
        self._cross20 = None
        self._w_arr = None
        self._bounding_circle = None

    def __eq__(self, other):
        tol = 1.0e-20

        if self._htmid == other._htmid:
            if self._level == other._level:
                if np.allclose(self._corners, other._corners, atol=tol):
                    return True

        return False

    def __ne__(self, other):
        return not (self == other)

    @property
    def htmid(self):
        """
        The unique integer identifying this trixel.
        """
        return self._htmid

    def contains(self, ra, dec):
        """
        Returns True if the specified RA, Dec are
        inside this trixel; False if not.

        RA and Dec are in degrees.
        """
        xyz = cartesian_from_spherical(np.radians(ra), np.radians(dec))
        return self.contains_pt(xyz)

    @property
    def cross01(self):
        """
        The cross product of the unit vectors defining
        the zeroth and first corners of this trixel.
        """
        if self._cross01 is None:
            self._cross01 = np.cross(self._corners[0], self._corners[1])
        return self._cross01

    @property
    def cross12(self):
        """
        The cross product of the unit vectors defining
        the first and second corners of this trixel.
        """
        if self._cross12 is None:
            self._cross12 = np.cross(self._corners[1], self._corners[2])
        return self._cross12

    @property
    def cross20(self):
        """
        The cross product of the unit vectors defining the second
        and zeroth corners of this trixel.
        """
        if self._cross20 is None:
            self._cross20 = np.cross(self._corners[2], self._corners[0])
        return self._cross20

    def _contains_one_pt(self, pt):
        """
        pt is a Cartesian point (not necessarily on the unit sphere).

        Returns True if the point projected onto the unit sphere
        is contained within this trixel; False if not.

        See equation 5 of

        Kunszt P., Szalay A., Thakar A. (2006) in "Mining The Sky",
        Banday A, Zaroubi S, Bartelmann M. eds.
        ESO Astrophysics Symposia
        https://www.researchgate.net/publication/226072008_The_Hierarchical_Triangular_Mesh
        """
        if np.dot(self.cross01, pt) >= 0.0:
            if np.dot(self.cross12, pt) >= 0.0:
                if np.dot(self.cross20, pt) >= 0.0:
                    return True

        return False

    def _contains_many_pts(self, pts):
        """
        pts is an array of Cartesian points (pts[0] is the zeroth
        point, pts[1] is the first point, etc.; not necessarily on
        the unit sphere).

        Returns an array of booleans denoting whether or not the
        projection of each point onto the unit sphere is contained
        within this trixel.

        See equation 5 of

        Kunszt P., Szalay A., Thakar A. (2006) in "Mining The Sky",
        Banday A, Zaroubi S, Bartelmann M. eds.
        ESO Astrophysics Symposia
        https://www.researchgate.net/publication/226072008_The_Hierarchical_Triangular_Mesh
        """
        return (
            (np.dot(pts, self.cross01) >= 0.0)
            & (np.dot(pts, self.cross12) >= 0.0)
            & (np.dot(pts, self.cross20) >= 0.0)
        )

    def contains_pt(self, pt):
        """
        pt is either a single Cartesian point
        or an array of Cartesian points (pt[0]
        is the zeroth point, pt[1] is the first
        point, etc.).

        Return a `bool` or array of booleans
        denoting whether this point(s) projected
        onto the unit sphere is/are contained within
        the current trixel.
        """
        if len(pt.shape) == 1:
            return self._contains_one_pt(pt)
        return self._contains_many_pts(pt)

    def _create_w(self):
        w0 = self._corners[1] + self._corners[2]
        w0 = w0 / np.sqrt(np.power(w0, 2).sum())
        w1 = self._corners[0] + self._corners[2]
        w1 = w1 / np.sqrt(np.power(w1, 2).sum())
        w2 = self._corners[0] + self._corners[1]
        w2 = w2 / np.sqrt(np.power(w2, 2).sum())

        self._w_arr = [w0, w1, w2]

    @property
    def w_arr(self):
        """
        An array of vectors needed to define the child trixels
        of this trixel.  See equation (3) of

        Kunszt P., Szalay A., Thakar A. (2006) in "Mining The Sky",
        Banday A, Zaroubi S, Bartelmann M. eds.
        ESO Astrophysics Symposia
        httpd://www.researchgate.net/publication/226072008_The_Hierarchical_Triangular_Mesh
        """
        if self._w_arr is None:
            self._create_w()
        return self._w_arr

    @property
    def t0(self):
        """
        The zeroth child trixel of this trixel.

        See Figure 2 of

        Szalay A. et al. (2007)
        "Indexing the Sphere with the Hierarchical Triangular Mesh"
        arXiv:cs/0701164
        """
        if not hasattr(self, "_t0"):
            self._t0 = Trixel(self.htmid << 2, [self._corners[0], self.w_arr[2], self.w_arr[1]])
        return self._t0

    @property
    def t1(self):
        """
        The first child trixel of this trixel.

        See Figure 2 of

        Szalay A. et al. (2007)
        "Indexing the Sphere with the Hierarchical Triangular Mesh"
        arXiv:cs/0701164
        """
        if not hasattr(self, "_t1"):
            self._t1 = Trixel((self.htmid << 2) + 1, [self._corners[1], self.w_arr[0], self.w_arr[2]])
        return self._t1

    @property
    def t2(self):
        """
        The second child trixel of this trixel.

        See Figure 2 of

        Szalay A. et al. (2007)
        "Indexing the Sphere with the Hierarchical Triangular Mesh"
        arXiv:cs/0701164
        """
        if not hasattr(self, "_t2"):
            self._t2 = Trixel((self.htmid << 2) + 2, [self._corners[2], self.w_arr[1], self.w_arr[0]])
        return self._t2

    @property
    def t3(self):
        """
        The third child trixel of this trixel.

        See Figure 2 of

        Szalay A. et al. (2007)
        "Indexing the Sphere with the Hierarchical Triangular Mesh"
        arXiv:cs/0701164
        """
        if not hasattr(self, "_t3"):
            self._t3 = Trixel((self.htmid << 2) + 3, [self.w_arr[0], self.w_arr[1], self.w_arr[2]])
        return self._t3

    def get_children(self):
        """
        Return a list of all of the child trixels of this trixel.
        """
        return [self.t0, self.t1, self.t2, self.t3]

    def get_child(self, dex):
        """
        Return a specific child trixel of this trixel.

        dex is an integer in the range [0,3] denoting
        which child to return

        See Figure 1 of

        Kunszt P., Szalay A., Thakar A. (2006) in "Mining The Sky",
        Banday A, Zaroubi S, Bartelmann M. eds.
        ESO Astrophysics Symposia
        https://www.researchgate.net/publication/226072008_The_Hierarchical_Triangular_Mesh

        for an explanation of which trixel corresponds to whic
        index.
        """
        if dex == 0:
            return self.t0
        elif dex == 1:
            return self.t1
        elif dex == 2:
            return self.t2
        elif dex == 3:
            return self.t3
        else:
            raise RuntimeError("Trixel has no %d child" % dex)

    def get_center(self):
        """
        Return the RA, Dec of the center of the circle bounding
        this trixel (RA, Dec both in degrees)
        """
        ra, dec = spherical_from_cartesian(self.bounding_circle[0])
        return np.degrees(ra), np.degrees(dec)

    def get_radius(self):
        """
        Return the angular radius in degrees of the circle bounding
        this trixel.
        """
        return np.degrees(self.bounding_circle[2])

    @property
    def level(self):
        """
        Return the level of subdivision for this trixel.  A higher
        level means a finer subdivision of the unit sphere and smaller
        trixels.  What we refer to as 'level' is denoted by 'd' in
        equation 2.5 of

        Szalay A. et al. (2007)
        "Indexing the Sphere with the Hierarchical Triangular Mesh"
        arXiv:cs/0701164

        For a given level == ell, there are 8*4**(ell-1) trixels in
        the entire unit sphere.

        The htmid values of trixels with level==ell will consist
        of 4 + 2*(ell-1) bits
        """
        return self._level

    @property
    def corners(self):
        """
        A numpy array containing the unit vectors pointing to the
        corners of this trixel.  corners[0] is the zeroth corner,
        corners[1] is the first corner, etc.
        """
        return self._corners

    @property
    def bounding_circle(self):
        """
        The circle on the unit sphere that bounds this trixel.

        See equation 4.2 of

        Szalay A. et al. (2007)
        "Indexing the Sphere with the Hierarchical Triangular Mesh"
        arXiv:cs/0701164

        Returns
        -------
        A tuple:
            Zeroth element is the unit vector pointing at
            the center of the bounding circle

            First element is the distance from the center of
            the unit sphere to the plane of the bounding circle
            (i.e. the dot product of the zeroth element with the
            most distant corner of the trixel).

            Second element is the half angular extent of the bounding circle.
        """
        if self._bounding_circle is None:
            # find the unit vector pointing to the center of the trixel
            vb = np.cross(
                (self._corners[1] - self._corners[0]),
                (self._corners[2] - self._corners[1]),
            )
            vb = vb / np.sqrt(np.power(vb, 2).sum())

            # find the distance from the center of the trixel
            # to the most distant corner of the trixel
            dd = np.dot(self.corners, vb).max()

            if np.abs(dd) > 1.0:
                raise RuntimeError("Bounding circle has dd %e (should be between -1 and 1)" % dd)

            self._bounding_circle = (vb, dd, np.arccos(dd))

        return self._bounding_circle


# Below are defined the initial Trixels
#
# See equations (1) and (2) of
#
# Kunszt P., Szalay A., Thakar A. (2006) in "Mining The Sky",
# Banday A, Zaroubi S, Bartelmann M. eds.
# ESO Astrophysics Symposia
# https://www.researchgate.net/publication/226072008_The_Hierarchical_Triangular_Mesh

_n0_trixel = Trixel(
    12,
    [np.array([1.0, 0.0, 0.0]), np.array([0.0, 0.0, 1.0]), np.array([0.0, -1.0, 0.0])],
)

_n1_trixel = Trixel(
    13,
    [np.array([0.0, -1.0, 0.0]), np.array([0.0, 0.0, 1.0]), np.array([-1.0, 0.0, 0.0])],
)

_n2_trixel = Trixel(
    14,
    [np.array([-1.0, 0.0, 0.0]), np.array([0.0, 0.0, 1.0]), np.array([0.0, 1.0, 0.0])],
)

_n3_trixel = Trixel(
    15,
    [np.array([0.0, 1.0, 0.0]), np.array([0.0, 0.0, 1.0]), np.array([1.0, 0.0, 0.0])],
)

_s0_trixel = Trixel(
    8,
    [np.array([1.0, 0.0, 0.0]), np.array([0.0, 0.0, -1.0]), np.array([0.0, 1.0, 0.0])],
)

_s1_trixel = Trixel(
    9,
    [np.array([0.0, 1.0, 0.0]), np.array([0.0, 0.0, -1.0]), np.array([-1.0, 0.0, 0.0])],
)

_s2_trixel = Trixel(
    10,
    [
        np.array([-1.0, 0.0, 0.0]),
        np.array([0.0, 0.0, -1.0]),
        np.array([0.0, -1.0, 0.0]),
    ],
)

_s3_trixel = Trixel(
    11,
    [np.array([0.0, -1.0, 0.0]), np.array([0.0, 0.0, -1.0]), np.array([1.0, 0.0, 0.0])],
)

basic_trixels = {
    "N0": _n0_trixel,
    "N1": _n1_trixel,
    "N2": _n2_trixel,
    "N3": _n3_trixel,
    "S0": _s0_trixel,
    "S1": _s1_trixel,
    "S2": _s2_trixel,
    "S3": _s3_trixel,
}


def level_from_htmid(htmid):
    """
    Find the level of a trixel from its htmid.  The level
    indicates how refined the triangular mesh is.

    There are 8*4**(d-1) triangles in a mesh of level=d

    (equation 2.5 of
    Szalay A. et al. (2007)
    "Indexing the Sphere with the Hierarchical Triangular Mesh"
    arXiv:cs/0701164)

    Note: valid htmids have 4+2n bits with a leading bit of 1
    """
    htmid_copy = htmid
    i_level = 1
    while htmid_copy > 15:
        htmid_copy >>= 2
        i_level += 1

    if htmid_copy < 8:
        raise RuntimeError(
            "\n%d is not a valid htmid.\n" % htmid
            + "Valid htmids will have 4+2n bits\n"
            + "with a leading bit of 1\n"
        )

    return i_level


def trixel_from_htmid(htmid):
    """
    Return the trixel corresponding to the given htmid
    (htmid is the unique integer identifying each trixel).

    Note: this method is not efficient for finding many
    trixels.  It recursively generates trixels and their
    children until it finds the right htmid without
    remembering which trixels it has already generated.
    To generate many trixels, use the get_all_trixels()
    method, which efficiently generates all of the trixels
    up to a given mesh level.

    Note: valid htmids have 4+2n bits with a leading bit of 1
    """
    level = level_from_htmid(htmid)
    base_htmid = htmid >> 2 * (level - 1)

    ans = None

    if base_htmid == 8:
        ans = _s0_trixel
    elif base_htmid == 9:
        ans = _s1_trixel
    elif base_htmid == 10:
        ans = _s2_trixel
    elif base_htmid == 11:
        ans = _s3_trixel
    elif base_htmid == 12:
        ans = _n0_trixel
    elif base_htmid == 13:
        ans = _n1_trixel
    elif base_htmid == 14:
        ans = _n2_trixel
    elif base_htmid == 15:
        ans = _n3_trixel

    if ans is None:
        raise RuntimeError("Unable to find trixel for id %d" % htmid)

    if level == 1:
        return ans

    # create an integer that is 4 bits
    # shorter than htmid (so it excludes
    # the bits corresponding to the base
    # trixel),with 11 in the two leading
    # positions
    complement = 3
    complement <<= 2 * (level - 2)

    for ix in range(level - 1):
        # Use bitwise and to figure out what the
        # two bits to the right of the bits
        # corresponding to the trixel currently
        # stored in ans are.  These two bits
        # determine which child of ans we need
        # to return.
        target = htmid & complement
        target >>= 2 * (level - ix - 2)

        if target >= 4:
            raise RuntimeError("target %d" % target)

        ans = ans.get_child(target)
        complement >>= 2

    return ans


def get_all_trixels(level):
    """
    Return a dict of all of the trixels up to a given mesh level.
    The dict is keyed on htmid, unique integer identifying
    each trixel on the unit sphere.  This method is efficient
    at generating many trixels at once.
    """

    # find how many bits should be added to the htmids
    # of the base trixels to get up to the desired level
    n_bits_added = 2 * (level - 1)

    # first, put the base trixels into the dict
    start_trixels = range(8, 16)
    trixel_dict = {}
    for t0 in start_trixels:
        trix0 = trixel_from_htmid(t0)
        trixel_dict[t0] = trix0

    ct = 0
    for t0 in start_trixels:
        t0 = t0 << n_bits_added
        for dt in range(2**n_bits_added):
            htmid = t0 + dt  # htmid we are currently generating
            ct += 1
            if htmid in trixel_dict:
                continue

            parent_id = htmid >> 2  # the immediate parent of htmid

            while parent_id not in trixel_dict:
                # find the highest-level ancestor trixel
                # of htmid that has already
                # been generated and added to trixel_dict
                for n_right in range(2, n_bits_added, 2):
                    if htmid >> n_right in trixel_dict:
                        break

                # generate the next highest level ancestor
                # of the current htmid
                to_gen = htmid >> n_right
                if to_gen in trixel_dict:
                    trix0 = trixel_dict[to_gen]
                else:
                    trix0 = trixel_from_htmid(to_gen)
                    trixel_dict[to_gen] = trix0

                # add the children of to_gen to trixel_dict
                trixel_dict[to_gen << 2] = trix0.get_child(0)
                trixel_dict[(to_gen << 2) + 1] = trix0.get_child(1)
                trixel_dict[(to_gen << 2) + 2] = trix0.get_child(2)
                trixel_dict[(to_gen << 2) + 3] = trix0.get_child(3)

            # add all of the children of parent_id to
            # trixel_dict
            trix0 = trixel_dict[parent_id]
            trixel_dict[(parent_id << 2)] = trix0.get_child(0)
            trixel_dict[(parent_id << 2) + 1] = trix0.get_child(1)
            trixel_dict[(parent_id << 2) + 2] = trix0.get_child(2)
            trixel_dict[(parent_id << 2) + 3] = trix0.get_child(3)

    return trixel_dict


def _iterate_trixel_finder(pt, parent, max_level):
    """
    Method to iteratively find the htmid of the trixel containing
    a point.

    Parameters
    ----------
    pt is a Cartesian point (not necessarily on the unit sphere)

    parent is the largest trixel currently known to contain the point

    max_level is the level of the triangular mesh at which we want
    to find the htmid.  Higher levels correspond to finer meshes.
    A mesh with level == ell contains 8*4**(ell-1) trixels.

    Returns
    -------
    The htmid at the desired level of the trixel containing the unit sphere
    projection of the point in pt.
    """
    children = parent.get_children()
    for child in children:
        if child.contains_pt(pt):
            if child.level == max_level:
                return child.htmid
            else:
                return _iterate_trixel_finder(pt, child, max_level)


def _find_htmid_slow(ra, dec, max_level):
    """
    Find the htmid (the unique integer identifying
    each trixel) of the trixel containing a given
    RA, Dec pair.

    Parameters
    ----------
    ra in degrees

    dec in degrees

    max_level is an integer denoting the mesh level
    of the trixel you want found

    Note: This method only works one point at a time.
    It cannot take arrays of RA and Dec.

    Returns
    -------
    An int (the htmid)
    """

    ra_rad = np.radians(ra)
    dec_rad = np.radians(dec)
    pt = cartesian_from_spherical(ra_rad, dec_rad)

    if _s0_trixel.contains_pt(pt):
        parent = _s0_trixel
    elif _s1_trixel.contains_pt(pt):
        parent = _s1_trixel
    elif _s2_trixel.contains_pt(pt):
        parent = _s2_trixel
    elif _s3_trixel.contains_pt(pt):
        parent = _s3_trixel
    elif _n0_trixel.contains_pt(pt):
        parent = _n0_trixel
    elif _n1_trixel.contains_pt(pt):
        parent = _n1_trixel
    elif _n2_trixel.contains_pt(pt):
        parent = _n2_trixel
    elif _n3_trixel.contains_pt(pt):
        parent = _n3_trixel
    else:
        raise RuntimeError("could not find parent Trixel")

    return _iterate_trixel_finder(pt, parent, max_level)


def _find_htmid_fast(ra, dec, max_level):
    """
    Find the htmid (the unique integer identifying
    each trixel) of the trixels containing arrays
    of RA, Dec pairs

    Parameters
    ----------
    ra in degrees (a numpy array)

    dec in degrees (a numpy array)

    max_level is an integer denoting the mesh level
    of the trixel you want found

    Returns
    -------
    A numpy array of ints (the htmids)

    Note: this method works by caching all of the trixels up to
    a given level.  Do not call it on max_level>10
    """

    if max_level > 10:
        raise RuntimeError(
            "Do not call _find_htmid_fast with max_level>10; "
            "the cache of trixels generated will be too large. "
            "Call find_htmid or _find_htmid_slow (find_htmid will "
            "redirect to _find_htmid_slow for large max_level)."
        )

    if not hasattr(_find_htmid_fast, "_trixel_dict") or _find_htmid_fast._level < max_level:
        _find_htmid_fast._trixel_dict = get_all_trixels(max_level)
        _find_htmid_fast._level = max_level

    ra_rad = np.radians(ra)
    dec_rad = np.radians(dec)
    pt_arr = cartesian_from_spherical(ra_rad, dec_rad)

    base_trixels = [
        _s0_trixel,
        _s1_trixel,
        _s2_trixel,
        _s3_trixel,
        _n0_trixel,
        _n1_trixel,
        _n2_trixel,
        _n3_trixel,
    ]

    htmid_arr = np.zeros(len(pt_arr), dtype=int)

    parent_dict = {}
    for parent in base_trixels:
        is_contained = parent.contains_pt(pt_arr)
        valid_dexes = np.where(is_contained)
        if len(valid_dexes[0]) == 0:
            continue
        htmid_arr[valid_dexes] = parent.htmid
        parent_dict[parent.htmid] = valid_dexes[0]

    for level in range(1, max_level):
        new_parent_dict = {}
        for parent_htmid in parent_dict.keys():
            considered_raw = parent_dict[parent_htmid]

            next_htmid = parent_htmid << 2
            children_htmid = [
                next_htmid,
                next_htmid + 1,
                next_htmid + 2,
                next_htmid + 3,
            ]

            is_found = np.zeros(len(considered_raw), dtype=int)
            for child in children_htmid:
                un_found = np.where(is_found == 0)[0]
                considered = considered_raw[un_found]
                if len(considered) == 0:
                    break
                child_trixel = _find_htmid_fast._trixel_dict[child]
                contains = child_trixel.contains_pt(pt_arr[considered])
                valid = np.where(contains)
                if len(valid[0]) == 0:
                    continue

                valid_dexes = considered[valid]
                is_found[un_found[valid[0]]] = 1
                htmid_arr[valid_dexes] = child
                new_parent_dict[child] = valid_dexes
        parent_dict = new_parent_dict

    return htmid_arr


def find_htmid(ra, dec, max_level):
    """
    Find the htmid (the unique integer identifying
    each trixel) of the trixel containing a given
    RA, Dec pair.

    Parameters
    ----------
    ra in degrees (either a number or a numpy array)

    dec in degrees (either a number or a numpy array)

    max_level is an integer denoting the mesh level
    of the trixel you want found

    Returns
    -------
    An int (the htmid) or an array of ints
    """
    if isinstance(ra, numbers.Number):
        are_arrays = False
    elif isinstance(ra, list):
        ra = np.array(ra)
        dec = np.array(dec)
        are_arrays = True
    else:
        try:
            assert isinstance(ra, np.ndarray)
            assert isinstance(dec, np.ndarray)
        except AssertionError:
            raise RuntimeError(
                "\nfindHtmid can handle types\n" + "RA: %s" % type(ra) + "Dec: %s" % type(dec) + "\n"
            )
        are_arrays = True

    if are_arrays:
        if max_level <= 10 and len(ra) > 100:
            return _find_htmid_fast(ra, dec, max_level)
        else:
            htmid_arr = np.zeros(len(ra), dtype=int)
            for ii in range(len(ra)):
                htmid_arr[ii] = _find_htmid_slow(ra[ii], dec[ii], max_level)
            return htmid_arr

    return _find_htmid_slow(ra, dec, max_level)


class HalfSpace:
    """
    HalfSpaces are circles on the unit sphere defined by intersecting
    a plane with the unit sphere.  They are specified by the unit vector
    pointing to their center on the unit sphere and the distance from
    the center of the unit sphere to the plane along that unit vector.

    See Section 3.1 of

    Szalay A. et al. (2007)
    "Indexing the Sphere with the Hierarchical Triangular Mesh"
    arXiv:cs/0701164

    Note that the specifying distance can be negative.  In this case,
    the halfspace is defined as the larger of the two regions on the
    unit sphere divided by the circle where the plane of the halfspace
    intersects the unit sphere.
    """

    def __init__(self, vector, length):
        """
        Parameters
        ----------
        vector is the unit vector pointing to the center of
        the halfspace on the unit sphere

        length is the distance from the center of the unit
        sphere to the plane defining the half space along the
        vector.  This length can be negative, in which case,
        the halfspace is defined as the larger of the two
        regions on the unit sphere divided by the circle
        where the plane of the halfspace intersects the
        unit sphere.
        """
        self._v = vector / np.sqrt(np.power(vector, 2).sum())
        self._d = length
        if np.abs(self._d) < 1.0:
            self._phi = np.arccos(self._d)  # half angular extent of the half space
            if self._phi > np.pi:
                raise RuntimeError("phi %e d %e" % (self._phi, self._d))
        else:
            if self._d < 0.0:
                self._phi = np.pi
            else:
                self._phi = 0.0

    def __eq__(self, other):
        tol = 1.0e-10
        if np.abs(self.dd - other.dd) > tol:
            return False
        if np.abs(np.dot(self.vector, other.vector) - 1.0) > tol:
            return False
        return True

    def __ne__(self, other):
        return not (self == other)

    @property
    def vector(self):
        """
        The unit vector from the origin to the center of the Half Space.
        """
        return self._v

    @property
    def dd(self):
        """
        The distance along the Half Space's vector that defines the
        extent of the Half Space.
        """
        return self._d

    @property
    def phi(self):
        """
        The angular radius of the Half Space on the surface of the sphere
        in radians.
        """
        return self._phi

    def contains_pt(self, pt, tol=None):
        """
        pt is a cartesian point (not necessarily on
        the unit sphere).  The method returns True if
        the projection of that point onto the unit sphere
        is contained in the halfspace; False otherwise.
        """
        norm_pt = pt / np.sqrt(np.power(pt, 2).sum())

        dot_product = np.dot(norm_pt, self._v)

        if tol is None:
            if dot_product > self._d:
                return True
        else:
            if dot_product > (self._d - tol):
                return True

        return False

    def contains_many_pts(self, pts):
        """
        Parameters
        ----------
        pts is a numpy array in which each row is a point on the
        unit sphere (note: must be normalized)

        Returns
        -------
        numpy array of booleans indicating which of pts are contained
        by this HalfSpace
        """
        dot_product = np.dot(pts, self._v)
        return dot_product > self._d

    def intersects_edge(self, pt1, pt2):
        """
        pt1 and pt2 are two unit vectors; the edge goes from pt1 to pt2.
        Return True if the edge intersects this halfspace; False otherwise.

        see equation 4.8 of

        Szalay A. et al. (2007)
        "Indexing the Sphere with the Hierarchical Triangular Mesh"
        arXiv:cs/0701164
        """
        costheta = np.dot(pt1, pt2)
        usq = (1 - costheta) / (1 + costheta)  # u**2; using trig identity for tan(theta/2)
        gamma1 = np.dot(self._v, pt1)
        gamma2 = np.dot(self._v, pt2)
        b = gamma1 * (usq - 1.0) + gamma2 * (usq + 1)
        a = -usq * (gamma1 + self._d)
        c = gamma1 - self._d

        det = b * b - 4 * a * c
        if det < 0.0:
            return False

        sqrt_det = np.sqrt(det)
        pos = (-b + sqrt_det) / (2.0 * a)

        if pos >= 0.0 and pos <= 1.0:
            return True

        neg = (-b - sqrt_det) / (2.0 * a)
        if neg >= 0.0 and neg <= 1.0:
            return True

        return False

    def intersects_circle(self, center, radius_rad):
        """
        Does this Half Space intersect a circle on the unit sphere

        center is the unit vector pointing to the center of the circle

        radius_rad is the radius of the circle in radians

        Returns a `bool`
        """

        dotproduct = np.dot(center, self._v)
        if np.abs(dotproduct) < 1.0:
            theta = np.arccos(dotproduct)
        elif dotproduct < 1.000000001 and dotproduct > 0.0:
            theta = 0.0
        elif dotproduct > -1.000000001 and dotproduct < 0.0:
            theta = np.pi
        else:
            raise RuntimeError("Dot product between unit vectors is %e" % dotproduct)

        if theta > self._phi + radius_rad:
            return False

        return True

    def intersects_bounding_circle(self, tx):
        """
        tx is a Trixel.  Return True if this halfspace intersects
        the bounding circle of the trixel; False otherwise.

        See the discussion around equation 4.2 of

        Szalay A. et al. (2007)
        "Indexing the Sphere with the Hierarchical Triangular Mesh"
        arXiv:cs/0701164
        """
        return self.intersects_circle(tx.bounding_circle[0], tx.bounding_circle[1])

    def contains_trixel(self, tx):
        """
        tx is a Trixel.

        Return "full" if the Trixel is fully contained by
        this halfspace.

        Return "partial" if the Trixel is only partially
        contained by this halfspace

        Return "outside" if no part of the Trixel is
        contained by this halfspace.

        See section 4.1 of

        Szalay A. et al. (2007)
        "Indexing the Sphere with the Hierarchical Triangular Mesh"
        arXiv:cs/0701164
        """

        containment = self.contains_many_pts(tx.corners)

        if containment.all():
            return "full"
        elif containment.any():
            return "partial"

        if tx.contains_pt(self._v):
            return "partial"

        # check if the trixel's bounding circle intersects
        # the halfspace
        if not self.intersects_bounding_circle(tx):
            return "outside"

        # need to test that the bounding circle intersect the halfspace
        # boundary

        for edge in (
            (tx.corners[0], tx.corners[1]),
            (tx.corners[1], tx.corners[2]),
            (tx.corners[2], tx.corners[0]),
        ):
            if self.intersects_edge(edge[0], edge[1]):
                return "partial"

        return "outside"

    @staticmethod
    def merge_trixel_bounds(bounds):
        """
        Take a list of trixel bounds as returned by HalfSpace.find_all_trixels
        and merge any tuples that should be merged

        Parameters
        ----------
        bounds is a list of trixel bounds as returned by HalfSpace.find_all_trixels

        Returns
        -------
        A new, equivalent list of trixel bounds
        """
        list_of_mins = np.array([r[0] for r in bounds])
        sorted_dex = np.argsort(list_of_mins)
        bounds_sorted = np.array(bounds)[sorted_dex]
        final_output = []
        current_list = []
        current_max = -1
        for row in bounds_sorted:
            if len(current_list) == 0 or row[0] <= current_max + 1:
                current_list.append(row[0])
                current_list.append(row[1])
                if row[1] > current_max:
                    current_max = row[1]
            else:
                final_output.append((min(current_list), max(current_list)))
                current_list = [row[0], row[1]]
                current_max = row[1]

        if len(current_list) > 0:
            final_output.append((min(current_list), max(current_list)))
        return final_output

    @staticmethod
    def join_trixel_bound_sets(b1, b2):
        """
        Take two sets of trixel bounds as returned by HalfSpace.find_all_trixels
        and return a set of trixel bounds that represents the intersection of
        the original sets of bounds
        """
        b1_sorted = HalfSpace.merge_trixel_bounds(b1)
        b2_sorted = HalfSpace.merge_trixel_bounds(b2)

        # maximum/minimum trixel bounds outside of which trixel ranges
        # will be considered invalid
        global_t_min = max(b1_sorted[0][0], b2_sorted[0][0])
        global_t_max = min(b1_sorted[-1][1], b2_sorted[-1][1])

        b1_keep = [r for r in b1_sorted if r[0] <= global_t_max and r[1] >= global_t_min]
        b2_keep = [r for r in b2_sorted if r[0] <= global_t_max and r[1] >= global_t_min]

        dex1 = 0
        dex2 = 0
        n_b1 = len(b1_keep)
        n_b2 = len(b2_keep)
        joint_bounds = []

        if n_b1 == 0 or n_b2 == 0:
            return joint_bounds

        while True:
            r1 = b1_keep[dex1]
            r2 = b2_keep[dex2]
            if r1[0] <= r2[0] and r1[1] >= r2[1]:
                # r2 is completely inside r1;
                # keep r2 and advance dex2
                joint_bounds.append(r2)
                dex2 += 1
            elif r2[0] <= r1[0] and r2[1] >= r1[1]:
                # r1 is completely inside r2;
                # keep r1 and advance dex1
                joint_bounds.append(r1)
                dex1 += 1
            else:
                # The two bounds are either disjoint, or they overlap;
                # find the intersection
                local_min = max(r1[0], r2[0])
                local_max = min(r1[1], r2[1])
                if local_min <= local_max:
                    # if we have a valid range, keep it
                    joint_bounds.append((local_min, local_max))

                # advance the bound that is lowest
                if r1[1] < r2[1]:
                    dex1 += 1
                else:
                    dex2 += 1

            # if we have finished scanning one or another of the
            # bounds, leave the loop
            if dex1 >= n_b1 or dex2 >= n_b2:
                break

        return HalfSpace.merge_trixel_bounds(joint_bounds)

    def find_all_trixels(self, level):
        """
        Find the HTMIDs of all of the trixels filling the half space

        Parameters
        ----------
        level is an integer denoting the resolution of the trixel grid

        Returns
        -------
        A list of tuples.  Each tuple gives an inclusive range of HTMIDs
        corresponding to trixels within the HalfSpace
        """

        global basic_trixels

        active_trixels = []
        for trixel_name in basic_trixels:
            active_trixels.append(basic_trixels[trixel_name])

        output_prelim = []
        max_d_htmid = 0

        # Once we establish that a given trixel is completely
        # contained within a the HalfSpace, we will need to
        # convert that trixel into a (min_htmid, max_htmid) pair.
        # This will involve evolving up from the current level
        # of trixel resolution to the desired level of trixel
        # resolution, setting the resulting 2-bit pairs to 0 for
        # min_htmid and 3 for max_htmid.  We can get min_htmid by
        # taking the base trixel's level and multiplying by an
        # appropriate power of 2.  We can get max_htmid by adding
        # an integer that, in binary, is wholly comprised of 1s
        # to min_htmid.  Here we construct that integer of 1s,
        # starting out at level-2, since the first trixels
        # to which we will need to add max_d_htmid will be
        # at least at level 2 (the children of the base trixels).
        for ii in range(level - 2):
            max_d_htmid += 3
            max_d_htmid <<= 2

        # start iterating at level 2 because level 1 is the base trixels,
        # where we are already starting, and i_level reallly refers to
        # the level of the child trixels we are investigating
        for i_level in range(2, level):
            max_d_htmid >>= 2

            new_active_trixels = []
            for tt in active_trixels:
                children = tt.get_children()
                for child in children:
                    is_contained = self.contains_trixel(child)
                    if is_contained == "partial":
                        # need to investigate more fully
                        new_active_trixels.append(child)
                    elif is_contained == "full":
                        # all of this trixels children, and their children are contained
                        min_htmid = child._htmid << 2 * (level - i_level)
                        max_htmid = min_htmid
                        max_htmid += max_d_htmid
                        output_prelim.append((min_htmid, max_htmid))

                        ########################################
                        # some assertions for debugging purposes
                        # assert min_htmid<max_htmid
                        # try:
                        #     test_trix = trixel_from_htmid(min_htmid)
                        #     assert self.contains_trixel(test_trix) != 'outside'
                        #     test_trix = trixel_from_htmid(max_htmid)
                        #     assert self.contains_trixel(test_trix) != 'outside'
                        # except AssertionError:
                        #     print('is_contained %s' % is_contained)
                        #     print('level %d' % level_from_htmid(tt._htmid))
                        #     raise

                active_trixels = new_active_trixels
            if len(active_trixels) == 0:
                break

        # final pass over the active_trixels to see which of their
        # children are inside this HalfSpace
        for trix in active_trixels:
            for child in trix.get_children():
                if self.contains_trixel(child) != "outside":
                    output_prelim.append((child._htmid, child._htmid))

        # sort output by htmid_min
        min_dex_arr = np.argsort([oo[0] for oo in output_prelim])
        output = []
        for ii in min_dex_arr:
            output.append(output_prelim[ii])

        return self.merge_trixel_bounds(output)


def half_space_from_ra_dec(ra, dec, radius):
    """
    Take an RA, Dec and radius of a circular field of view and return
    a HalfSpace

    Parameters
    ----------
    ra in degrees

    dec in degrees

    radius in degrees

    Returns
    -------
    HalfSpace corresponding to the circular field of view
    """
    dd = np.cos(np.radians(radius))
    xyz = cartesian_from_spherical(np.radians(ra), np.radians(dec))
    return HalfSpace(xyz, dd)


def half_space_from_points(pt1, pt2, pt3):
    """
    Return a Half Space defined by two points on a Great Circle
    and a third point contained in the Half Space.

    Parameters
    ----------
    pt1, pt2 -- two tuples containing (RA, Dec) in degrees of
    the points on the Great Circle defining the Half Space

    pt3 -- a tuple containing (RA, Dec) in degrees of a point
    contained in the Half Space

    Returns
    --------
    A Half Space
    """

    vv1 = cartesian_from_spherical(np.radians(pt1[0]), np.radians(pt1[1]))
    vv2 = cartesian_from_spherical(np.radians(pt2[0]), np.radians(pt2[1]))
    axis = np.array(
        [
            vv1[1] * vv2[2] - vv1[2] * vv2[1],
            vv1[2] * vv2[0] - vv1[0] * vv2[2],
            vv1[0] * vv2[1] - vv1[1] * vv2[0],
        ]
    )

    axis /= np.sqrt(np.sum(axis**2))

    vv3 = cartesian_from_spherical(np.radians(pt3[0]), np.radians(pt3[1]))
    if np.dot(axis, vv3) < 0.0:
        axis *= -1.0

    return HalfSpace(axis, 0.0)


def intersect_half_spaces(hs1, hs2):
    """
    Parameters
    ----------
    hs1, hs2 are Half Spaces

    Returns
    -------
    A list of the cartesian points where the Half Spaces intersect.

    Note: if the Half Spaces are identical, then this list will be
    empty.

    Based on section 3.5 of
    Szalay A. et al. (2007)
    "Indexing the Sphere with the Hierarchical Triangular Mesh"
    arXiv:cs/0701164
    """
    gamma = np.dot(hs1.vector, hs2.vector)
    if np.abs(1.0 - np.abs(gamma)) < 1.0e-20:
        # Half Spaces are based on parallel planes that don't intersect
        return []

    denom = 1.0 - gamma**2
    if denom < 0.0:
        return []

    num = hs1.dd**2 + hs2.dd**2 - 2.0 * gamma * hs1.dd * hs2.dd
    if denom < num:
        return []

    uu = (hs1.dd - gamma * hs2.dd) / denom
    vv = (hs2.dd - gamma * hs1.dd) / denom

    ww = np.sqrt((1.0 - num / denom) / denom)
    cross_product = np.array(
        [
            hs1.vector[1] * hs2.vector[2] - hs1.vector[2] * hs2.vector[1],
            hs1.vector[2] * hs2.vector[0] - hs1.vector[0] * hs2.vector[2],
            hs1.vector[0] * hs2.vector[1] - hs1.vector[1] * hs2.vector[0],
        ]
    )

    pt1 = uu * hs1.vector + vv * hs2.vector + ww * cross_product
    pt2 = uu * hs1.vector + vv * hs2.vector - ww * cross_product
    if np.abs(1.0 - np.dot(pt1, pt2)) < 1.0e-20:
        return pt1
    return np.array([pt1, pt2])



import os
import unittest

from astropy.time import Time, TimeDelta

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.site_models import SeeingData

# Unit test here uses oldest/original opsim seeing database, "Seeing.db".
# Could be updated to use a new DB, but that would require changing some of these hard-coded numbers.


class TestSeeingData(unittest.TestCase):
    def setUp(self):
        self.time = Time("2020-01-01", format="isot", scale="tai")
        self.seeing_db = os.path.join(get_data_dir(), "site_models", "seeing.db")
        print(self.seeing_db)

    def test_information_after_read(self):
        seeing_data = SeeingData(self.time, seeing_db=self.seeing_db)
        seeing_data.read_data()
        self.assertTrue(len(seeing_data.seeing_values) > 0)
        self.assertTrue(len(seeing_data.seeing_dates) > 0)
        self.assertEqual(seeing_data.start_time, self.time)
        self.assertEqual(seeing_data.seeing_db, self.seeing_db)

    def test_fwhm500_at_time(self):
        seeing_data = SeeingData(self.time, self.seeing_db, offset_year=0)
        seeing_data.read_data()
        dt = TimeDelta(75400, format="sec")
        self.assertEqual(seeing_data(self.time + dt), 0.859431982040405)
        dt = TimeDelta(76700, format="sec")
        self.assertEqual(seeing_data(self.time + dt), 0.646009027957916)
        dt = TimeDelta(63190400, format="sec")
        self.assertEqual(seeing_data(self.time + dt), 0.64860999584198)
        dt = TimeDelta(189424900, format="sec")
        self.assertEqual(seeing_data(self.time + dt), 0.699440002441406)
        # Test time selection from seeing data.
        dt = TimeDelta(800, format="sec")
        fwhm500 = seeing_data(self.time + dt)
        # Hack seeing data to remove first date, thus db does not start at zero.
        seeing_data.seeing_dates = seeing_data.seeing_dates[:-1]
        seeing_data.seeing_values = seeing_data.seeing_values[:-1]
        seeing_data.time_range = seeing_data.seeing_dates[-1] - seeing_data.seeing_dates[0]
        seeing_data.min_time = seeing_data.seeing_dates[0]
        self.assertEqual(fwhm500, seeing_data(self.time + dt))

    def test_using_different_start_month(self):
        t2 = Time("2020-05-24", format="isot", scale="tai")
        seeing_data = SeeingData(t2, self.seeing_db, offset_year=0)
        self.assertEqual(seeing_data.start_time, self.time)
        seeing_data.read_data()
        dt = TimeDelta(75400, format="sec")
        self.assertEqual(seeing_data(t2 + dt), 0.437314003705978)
        dt = TimeDelta(63190400, format="sec")
        self.assertEqual(seeing_data(t2 + dt), 0.453994989395142)


if __name__ == "__main__":
    unittest.main()



__all__ = ("sims_clean_up",)

import gc
import numbers

import numpy as np


def sims_clean_up():
    """
    This method will clean up data caches created by the sims software stack.
    Any time a cache is added to the sims software stack, it can be added to
    the list sims_clean_up.targets.  When sims_clean_up() is called, it will
    loop through the contents of sims_clean_up.targets.  It will call pop()
    on all of the contents of each sims_clean_up.target, run close() on each
    item it pops (if applicable), delete each item it pops, and then reset
    each sims_clean_up.target to either a blank dict or list (depending on
    what the target was).  sims_clean_up() will then run the garbage
    collector.

    Note: if a target cache is not a dict or a list, it will attempt to call
    close() on the cache and delete the cache.
    """

    if not hasattr(sims_clean_up, "targets"):
        return None

    for target in sims_clean_up.targets:
        if isinstance(target, dict):
            while len(target) > 0:
                obj = target.popitem()
                if hasattr(obj[1], "close"):
                    try:
                        obj[1].close()
                    except AttributeError:
                        pass
                del obj
        elif isinstance(target, list):
            while len(target) > 0:
                obj = target.pop()
                if hasattr(obj, "close"):
                    try:
                        obj.close()
                    except AttributeError:
                        pass
                del obj
        else:
            if hasattr(target, "close"):
                target.close()
            del target

    gc.collect()
    return None


sims_clean_up.targets = []


def _validate_inputs(input_list, input_names, method_name):
    """
    This method will validate the inputs of other methods.

    input_list is a list of the inputs passed to a method.

    input_name is a list of the variable names associated with
    input_list

    method_name is the name of the method whose input is being validated.

    _validate_inputs will verify that all of the inputs in input_list are:

    1) of the same type
    2) either numpy arrays or instances of numbers.Number (floats or ints)
    3) if they are numpy arrays, they all have the same length

    If any of these criteria are violated, a RuntimeError will be raised

    returns True if the inputs are numpy arrays; False if not
    """

    if isinstance(input_list[0], np.ndarray):
        desired_type = np.ndarray
    elif isinstance(input_list[0], numbers.Number):
        desired_type = numbers.Number
    else:
        raise RuntimeError(
            "The arg %s input to method %s " % (input_names[0], method_name)
            + "should be either a number or a numpy array"
        )

    valid_type = True
    bad_names = []
    for ii, nn in zip(input_list, input_names):
        if not isinstance(ii, desired_type):
            valid_type = False
            bad_names.append(nn)

    if not valid_type:
        msg = "The input arguments:\n"
        for nn in bad_names:
            msg += "%s,\n" % nn
        msg += "passed to %s " % method_name
        msg += "need to be either numbers or numpy arrays\n"
        msg += "and the same type as the argument %s" % input_names[0]
        msg += "\n\nTypes of arguments are:\n"
        for name, arg in zip(input_names, input_list):
            msg += "%s: %s\n" % (name, type(arg))
        raise RuntimeError(msg)

    if desired_type is np.ndarray:
        same_length = True
        for ii in input_list:
            if len(ii) != len(input_list[0]):
                same_length = False
        if not same_length:
            raise RuntimeError("The arrays input to %s " % method_name + "all need to have the same length")

    if desired_type is np.ndarray:
        return True

    return False



from .approx_coord_transforms import *
from .bearing import *
from .binned_stats import *
from .camera_footprint import *
from .coordinate_transformations import *
from .ddf_locations import *
from .file_maps import *
from .get_package_dir import *
from .healpy_utils import *
from .htm_module import *
from .m5_flat_sed import *
from .mjd_zero import *
from .projections import *
from .sampling_functions import *
from .season_utils import *
from .site import *
from .sys_eng_vals import *
from .tree_utils import *
from .zernike_module import *



__all__ = ("sim_runner",)

import sqlite3
import sys
import time
import warnings

import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.schedulers import SimpleFilterSched
from rubin_scheduler.scheduler.utils import SchemaConverter, empty_observation, run_info_table


def sim_runner(
    observatory,
    scheduler,
    filter_scheduler=None,
    mjd_start=None,
    survey_length=3.0,
    filename=None,
    delete_past=True,
    n_visit_limit=None,
    step_none=15.0,
    verbose=True,
    extra_info=None,
    event_table=None,
    record_rewards=False,
    start_result_size=int(2e5),
    append_result_size=int(2.5e6),
):
    """
    run a simulation

    Parameters
    ----------
    survey_length : float (3.)
        The length of the survey ot run (days)
    step_none : float (15)
        The amount of time to advance if the scheduler fails to return a target (minutes).
    extra_info : dict (None)
        If present, dict gets added onto the information from the observatory model.
    event_table : np.array (None)
        Any ToO events that were included in the simulation
    record_rewards : bool (False)
        Save computed rewards
    start_result_size : int
        Size of observations array to pre-allocate at the start of the run. Default 2e5.
    append_result_size : int
        Size of observations array to append if start_result_size is too small.
        Default 2.5e6.
    """

    if extra_info is None:
        extra_info = {}

    t0 = time.time()

    if filter_scheduler is None:
        filter_scheduler = SimpleFilterSched()

    if mjd_start is None:
        mjd = observatory.mjd + 0
        mjd_start = mjd + 0
    else:
        mjd = mjd_start + 0
        observatory.mjd = mjd

    end_mjd = mjd + survey_length
    observations = empty_observation(n=start_result_size)
    mjd_track = mjd + 0
    step = 1.0 / 24.0
    step_none = step_none / 60.0 / 24.0  # to days
    mjd_run = end_mjd - mjd_start
    nskip = 0
    new_night = False

    mjd_last_flush = -1

    last_obs_queue_fill_mjd_ns = None
    obs_rewards = {}
    reward_dfs = []

    counter = 0
    while mjd < end_mjd:
        if not scheduler._check_queue_mjd_only(observatory.mjd):
            scheduler.update_conditions(observatory.return_conditions())
        desired_obs = scheduler.request_observation(mjd=observatory.mjd)
        if record_rewards:
            if last_obs_queue_fill_mjd_ns != scheduler.queue_fill_mjd_ns:
                reward_dfs.append(scheduler.queue_reward_df)
                last_obs_queue_fill_mjd_ns = scheduler.queue_fill_mjd_ns

        if desired_obs is None:
            # No observation. Just step into the future and try again.
            warnings.warn("No observation. Step into the future and trying again.")
            observatory.mjd = observatory.mjd + step_none
            scheduler.update_conditions(observatory.return_conditions())
            nskip += 1
            continue
        completed_obs, new_night = observatory.observe(desired_obs)
        if completed_obs is not None:
            scheduler.add_observation(completed_obs[0])
            observations[counter] = completed_obs[0]
            filter_scheduler.add_observation(completed_obs[0])
            counter += 1
            if counter == observations.size:
                add_observations = empty_observation(n=append_result_size)
                observations = np.concatenate([observations, add_observations])

            if record_rewards:
                obs_rewards[completed_obs[0]["mjd"]] = last_obs_queue_fill_mjd_ns
        else:
            # An observation failed to execute, usually it was outside the altitude limits.
            if observatory.mjd == mjd_last_flush:
                raise RuntimeError("Scheduler has failed to provide a valid observation multiple times.")
            # if this is a first offence, might just be that targets set.
            # Flush queue and get some new targets.
            scheduler.flush_queue()
            mjd_last_flush = observatory.mjd + 0
        if new_night:
            # find out what filters we want mounted
            conditions = observatory.return_conditions()
            filters_needed = filter_scheduler(conditions)
            observatory.observatory.mount_filters(filters_needed)

        mjd = observatory.mjd + 0
        if verbose:
            if (mjd - mjd_track) > step:
                progress = np.max((mjd - mjd_start) / mjd_run * 100)
                text = "\rprogress = %.2f%%" % progress
                sys.stdout.write(text)
                sys.stdout.flush()
                mjd_track = mjd + 0
        if n_visit_limit is not None:
            if counter == n_visit_limit:
                break
        # XXX--handy place to interupt and debug
        # if len(observations) > 25:
        #    import pdb ; pdb.set_trace()

    # trim off any observations that were pre-allocated but not used
    observations = observations[0:counter]

    runtime = time.time() - t0
    print("Skipped %i observations" % nskip)
    print("Flushed %i observations from queue for being stale" % scheduler.flushed)
    print("Completed %i observations" % len(observations))
    print("ran in %i min = %.1f hours" % (runtime / 60.0, runtime / 3600.0))
    if len(observations) > 0:
        if filename is not None:
            print("Writing results to ", filename)
        if filename is not None:
            info = run_info_table(observatory, extra_info=extra_info)
            converter = SchemaConverter()
            converter.obs2opsim(observations, filename=filename, info=info, delete_past=delete_past)
        if event_table is not None:
            df = pd.DataFrame(event_table)
            con = sqlite3.connect(filename)
            df.to_sql("events", con)
            con.close()

        if record_rewards:
            reward_df = pd.concat(reward_dfs)
            obs_rewards_series = pd.Series(obs_rewards)
            obs_rewards_series.index.name = "mjd"
            obs_rewards_series.name = "queue_fill_mjd_ns"

            if filename is not None:
                with sqlite3.connect(filename) as con:
                    reward_df.to_sql("rewards", con)
                    obs_rewards_series.to_sql("obs_rewards", con)
    else:
        # Make sure there is something to return if there are no observations
        reward_df = None
        obs_rewards_series = None

    if record_rewards:
        result = observatory, scheduler, observations, reward_df, obs_rewards_series
    else:
        result = observatory, scheduler, observations

    return result



__all__ = ("generate_ddf_scheduled_obs",)

import os
import warnings

import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler.utils import scheduled_observation
from rubin_scheduler.utils import calc_season, ddf_locations, survey_start_mjd


def ddf_slopes(ddf_name, raw_obs, night_season):
    """
    Let's make custom slopes for each DDF

    Parameters
    ----------
    ddf_name : str
       The DDF name to use
    raw_obs : np.array
        An array with values of 1 or zero. One element per night, value of
        1 indicates the night is during an active observing season.
    night_season : np.array
        An array of floats with the fractional season value
        (e.g., 0.5 would be half way through the first season)
    """

    # OK, so 258 sequences is ~1% of the survey
    # so a 25.8 sequences is a 0.1% season
    # COSMOS is going to be 0.7% for 3 years, then 0.175 for the rest.

    ss = 30  # standard season, was 45

    if (ddf_name == "ELAISS1") | (ddf_name == "XMM_LSS") | (ddf_name == "ECDFS"):
        # Dict with keys for each season and values of the number of sequences
        # to attempt.
        season_vals = {
            0: 10,
            1: ss,
            2: ss,
            3: ss,
            4: ss,
            5: ss,
            6: ss,
            7: ss,
            8: ss,
            9: ss,
            10: 10,
        }

        round_season = np.floor(night_season)

        cumulative_desired = np.zeros(raw_obs.size, dtype=float)
        for season in np.unique(round_season):
            in_season = np.where(round_season == season)
            cumulative = np.cumsum(raw_obs[in_season])
            if cumulative.max() > 0:
                cumulative = cumulative / cumulative.max() * season_vals[season]
                cumulative_desired[in_season] = cumulative + np.max(cumulative_desired)

    if ddf_name == "EDFS_a":
        season_vals = {
            0: 10,
            1: ss,
            2: ss,
            3: ss,
            4: ss,
            5: ss,
            6: ss,
            7: ss,
            8: ss,
            9: ss,
            10: 10,
        }

        round_season = np.floor(night_season)

        cumulative_desired = np.zeros(raw_obs.size, dtype=float)
        for season in np.unique(round_season):
            in_season = np.where(round_season == season)
            cumulative = np.cumsum(raw_obs[in_season])
            if cumulative.max() > 0:
                cumulative = cumulative / cumulative.max() * season_vals[season]
                cumulative_desired[in_season] = cumulative + np.max(cumulative_desired)

    if ddf_name == "COSMOS":
        # looks like COSMOS has no in-season time for 10 at the current start mjd.
        season_vals = {
            0: 10,
            1: ss * 5,
            2: ss * 5,
            3: ss * 2,
            4: ss,
            5: ss,
            6: ss,
            7: ss,
            8: ss,
            9: ss,
            10: 10,
        }

        round_season = np.floor(night_season)

        cumulative_desired = np.zeros(raw_obs.size, dtype=float)
        for season in np.unique(round_season):
            in_season = np.where(round_season == season)[0]
            cumulative = np.cumsum(raw_obs[in_season])
            if cumulative.max() > 0:
                cumulative = cumulative / cumulative.max() * season_vals[season]
                cumulative_desired[in_season] = cumulative + np.max(cumulative_desired)

    return cumulative_desired


def match_cumulative(cumulative_desired, mask=None, no_duplicate=True):
    """Generate a schedule that tries to match the desired cumulative distribution given a mask

    Parameters
    ----------
    cumulative_desired : `np.array`, float
        An array with the cumulative number of desired observations. Elements
        are assumed to be evenly spaced.
    mask : `np.array`, bool or int (None)
        Set to zero for indices that cannot be scheduled
    no_duplicate : `bool` (True)
        If True, only 1 event can be scheduled per element

    Returns
    -------
    schedule : `np.array`
        The resulting schedule, with values marking number of events in that cell.
    """

    rounded_desired = np.round(cumulative_desired)
    sched = cumulative_desired * 0
    if mask is None:
        mask = np.ones(sched.size)

    valid = np.where(mask > 0)[0].tolist()
    x = np.arange(sched.size)

    drd = np.diff(rounded_desired)
    step_points = np.where(drd > 0)[0] + 1

    # would be nice to eliminate this loop, but it's not too bad.
    # can't just use searchsorted on the whole array, because then there
    # can be duplicate values, and array[[n,n]] = 1 means that extra match gets lost.
    for indx in step_points:
        left = np.searchsorted(x[valid], indx)
        right = np.searchsorted(x[valid], indx, side="right")
        d1 = indx - left
        d2 = right - indx
        if d1 < d2:
            sched_at = left
        else:
            sched_at = right

        # If we are off the end
        if sched_at >= len(valid):
            sched_at -= 1

        sched[valid[sched_at]] += 1
        if no_duplicate:
            valid.pop(sched_at)

    return sched


def optimize_ddf_times(
    ddf_name,
    ddf_RA,
    ddf_grid,
    sun_limit=-18,
    sequence_time=60.0,
    airmass_limit=2.5,
    sky_limit=None,
    g_depth_limit=23.5,
    season_unobs_frac=0.1,
):
    """

    Parameters
    ----------
    ddf : `str`
        The name of the DDF
    ddf_grid : `np.array`
        An array with info for the DDFs. Generated by the
        rubin_scheduler.scheduler/surveys/generate_ddf_grid.py` script
    season_unobs_frac : `float`
        7.2 month observing season if season_unobs_frac = 0.2 (shaves 20% off
        each end of the full year)
    sequence_time : `float`
        How long a sequence is expected to be (minutes). Used to make
        sure things are not scheduled too close to twilight.
    """
    sun_limit = np.radians(sun_limit)
    sequence_time = sequence_time / 60.0 / 24.0  # to days

    # XXX-- double check that I got this right
    ack = ddf_grid["sun_alt"][0:-1] * ddf_grid["sun_alt"][1:]
    night = np.zeros(ddf_grid.size, dtype=int)
    night[np.where((ddf_grid["sun_alt"][1:] >= 0) & (ack < 0))] += 1
    night = np.cumsum(night)
    ngrid = ddf_grid["mjd"].size

    # set a sun, airmass, sky masks
    sun_mask = np.ones(ngrid, dtype=int)
    sun_mask[np.where(ddf_grid["sun_alt"] >= sun_limit)] = 0

    # expand sun mask backwards by the sequence time.
    n_back = np.ceil(sequence_time / (ddf_grid["mjd"][1] - ddf_grid["mjd"][0])).astype(int)
    shadow_indx = np.where(sun_mask == 0)[0] - n_back
    shadow_indx = shadow_indx[np.where(shadow_indx >= 0)]

    sun_mask[shadow_indx] = 0

    airmass_mask = np.ones(ngrid, dtype=int)
    airmass_mask[np.where(ddf_grid["%s_airmass" % ddf_name] >= airmass_limit)] = 0

    sky_mask = np.ones(ngrid, dtype=int)
    if sky_limit is not None:
        sky_mask[np.where(ddf_grid["%s_sky_g" % ddf_name] <= sky_limit)] = 0
        sky_mask[np.where(np.isnan(ddf_grid["%s_sky_g" % ddf_name]) == True)] = 0

    m5_mask = np.zeros(ngrid, dtype=bool)
    m5_mask[np.isfinite(ddf_grid["%s_m5_g" % ddf_name])] = 1

    if g_depth_limit is not None:
        m5_mask[np.where(ddf_grid["%s_m5_g" % ddf_name] < g_depth_limit)] = 0

    big_mask = sun_mask * airmass_mask * sky_mask * m5_mask

    potential_nights = np.unique(night[np.where(big_mask > 0)])

    # prevent a repeat sequence in a night
    unights, indx = np.unique(night, return_index=True)
    night_mjd = ddf_grid["mjd"][indx]
    # The season of each night
    night_season = calc_season(ddf_RA, night_mjd)

    raw_obs = np.ones(unights.size)
    # take out the ones that are out of season
    season_mod = night_season % 1

    out_season = np.where((season_mod < season_unobs_frac) | (season_mod > (1.0 - season_unobs_frac)))
    raw_obs[out_season] = 0

    cumulative_desired = ddf_slopes(ddf_name, raw_obs, night_season)

    night_mask = unights * 0
    night_mask[potential_nights] = 1

    unight_sched = match_cumulative(cumulative_desired, mask=night_mask)
    cumulative_sched = np.cumsum(unight_sched)

    nights_to_use = unights[np.where(unight_sched == 1)]

    # For each night, find the best time in the night.
    # XXX--probably need to expand this part to resolve the times when multiple things get scheduled
    mjds = []
    for night_check in nights_to_use:
        in_night = np.where((night == night_check) & (np.isfinite(ddf_grid["%s_m5_g" % ddf_name])))[0]
        m5s = ddf_grid["%s_m5_g" % ddf_name][in_night]
        # we could intorpolate this to get even better than 15 min resolution on when to observe
        max_indx = np.where(m5s == m5s.max())[0].min()
        mjds.append(ddf_grid["mjd"][in_night[max_indx]])

    return mjds, night_mjd, cumulative_desired, cumulative_sched


def generate_ddf_scheduled_obs(
    data_file=None,
    flush_length=2,
    mjd_tol=15,
    expt=30.0,
    alt_min=25,
    alt_max=85,
    HA_min=21.0,
    HA_max=3.0,
    sun_alt_max=-18,
    dist_tol=3.0,
    season_unobs_frac=0.1,
    nvis_master=[8, 10, 20, 20, 24, 18],
    filters="ugrizy",
    nsnaps=[1, 2, 2, 2, 2, 2],
    mjd_start=None,
    survey_length=10.0,
    sequence_time=60.0,
):
    """

    Parameters
    ----------
    data_file : `path` (None)
        The data file to use for DDF airmass, m5, etc. Defaults to using whatever is in
        rubin_sim_data/scheduler directory.
    flush_length : `float` (2)
        How long to keep a scheduled observation around before it is considered failed
        and flushed (days).
    mjd_tol : `float` (15)
        How close an observation must be in time to be considered matching a scheduled
        observation (minutes).
    expt : `float` (30)
        Total exposure time per visit (seconds).
    alt_min/max : `float` (25, 85)
        The minimum and maximum altitudes to permit observations to happen (degrees).
    HA_min/max : `float` (21, 3)
        The hour angle limits to permit observations to happen (hours).
    dist_tol : `float` (3)
        The distance tolerance for a visit to be considered matching a scheduled observation
        (degrees).
    season_unobs_frac : `float` (0.1)
        What fraction of the season should the DDF be considered unobservable. Taken off both the
        start and end of the year, so a season frac of 0.1 means 20% of the time the DDF is considered
        unobservable, so it will be in-season for 9.6 months.
    nvis_master : list of ints ([8, 10, 20, 20, 24, 18])
        The number of visits to make per filter
    filters : `str` (ugrizy)
        The filter names.
    nsnaps : `list of ints` ([1, 2, 2, 2, 2, 2])
        The number of snaps to use per filter
    mjd_start : `float`
        Starting MJD of the survey. Default None, which calls rubin_sim.utils.survey_start_mjd
    survey_length : `float`
        Length of survey (years). Default 10.
    """
    if data_file is None:
        data_file = os.path.join(get_data_dir(), "scheduler", "ddf_grid.npz")

    if mjd_start is None:
        mjd_start = survey_start_mjd()

    flush_length = flush_length  # days
    mjd_tol = mjd_tol / 60 / 24.0  # minutes to days
    expt = expt
    alt_min = np.radians(alt_min)
    alt_max = np.radians(alt_max)
    dist_tol = np.radians(dist_tol)
    sun_alt_max = np.radians(sun_alt_max)

    ddfs = ddf_locations()
    ddf_data = np.load(data_file)
    ddf_grid = ddf_data["ddf_grid"].copy()

    mjd_max = mjd_start + survey_length * 365.25

    # check if our pre-computed grid is over the time range we think we are scheduling for
    if (ddf_grid["mjd"].min() > mjd_start) | (ddf_grid["mjd"].max() < mjd_max):
        warnings.warn("Pre-computed DDF properties don't match requested survey times")

    in_range = np.where((ddf_grid["mjd"] >= mjd_start) & (ddf_grid["mjd"] <= mjd_max))
    ddf_grid = ddf_grid[in_range]

    all_scheduled_obs = []
    for ddf_name in ["ELAISS1", "XMM_LSS", "ECDFS", "COSMOS", "EDFS_a"]:
        print("Optimizing %s" % ddf_name)

        # 'ID', 'RA', 'dec', 'mjd', 'flush_by_mjd', 'exptime', 'filter', 'rotSkyPos', 'nexp',
        #         'note'
        # 'mjd_tol', 'dist_tol', 'alt_min', 'alt_max', 'HA_max', 'HA_min', 'observed'
        mjds = optimize_ddf_times(
            ddf_name,
            ddfs[ddf_name][0],
            ddf_grid,
            season_unobs_frac=season_unobs_frac,
            sequence_time=sequence_time,
        )[0]
        for mjd in mjds:
            for filtername, nvis, nexp in zip(filters, nvis_master, nsnaps):
                if "EDFS" in ddf_name:
                    obs = scheduled_observation(n=int(nvis / 2))
                    obs["RA"] = np.radians(ddfs[ddf_name][0])
                    obs["dec"] = np.radians(ddfs[ddf_name][1])
                    obs["mjd"] = mjd
                    obs["flush_by_mjd"] = mjd + flush_length
                    obs["exptime"] = expt
                    obs["filter"] = filtername
                    obs["nexp"] = nexp
                    obs["note"] = "DD:%s" % ddf_name
                    obs["target"] = ddf_name

                    obs["mjd_tol"] = mjd_tol
                    obs["dist_tol"] = dist_tol
                    # Need to set something for HA limits
                    obs["HA_min"] = HA_min
                    obs["HA_max"] = HA_max
                    obs["alt_min"] = alt_min
                    obs["alt_max"] = alt_max
                    obs["sun_alt_max"] = sun_alt_max
                    all_scheduled_obs.append(obs)

                    obs = scheduled_observation(n=int(nvis / 2))
                    obs["RA"] = np.radians(ddfs[ddf_name.replace("_a", "_b")][0])
                    obs["dec"] = np.radians(ddfs[ddf_name.replace("_a", "_b")][1])
                    obs["mjd"] = mjd
                    obs["flush_by_mjd"] = mjd + flush_length
                    obs["exptime"] = expt
                    obs["filter"] = filtername
                    obs["nexp"] = nexp
                    obs["note"] = "DD:%s" % ddf_name.replace("_a", "_b")
                    obs["target"] = ddf_name.replace("_a", "_b")

                    obs["mjd_tol"] = mjd_tol
                    obs["dist_tol"] = dist_tol
                    # Need to set something for HA limits
                    obs["HA_min"] = HA_min
                    obs["HA_max"] = HA_max
                    obs["alt_min"] = alt_min
                    obs["alt_max"] = alt_max
                    obs["sun_alt_max"] = sun_alt_max
                    all_scheduled_obs.append(obs)

                else:
                    obs = scheduled_observation(n=nvis)
                    obs["RA"] = np.radians(ddfs[ddf_name][0])
                    obs["dec"] = np.radians(ddfs[ddf_name][1])
                    obs["mjd"] = mjd
                    obs["flush_by_mjd"] = mjd + flush_length
                    obs["exptime"] = expt
                    obs["filter"] = filtername
                    obs["nexp"] = nexp
                    obs["note"] = "DD:%s" % ddf_name
                    obs["target"] = ddf_name

                    obs["mjd_tol"] = mjd_tol
                    obs["dist_tol"] = dist_tol
                    # Need to set something for HA limits
                    obs["HA_min"] = HA_min
                    obs["HA_max"] = HA_max
                    obs["alt_min"] = alt_min
                    obs["alt_max"] = alt_max
                    obs["sun_alt_max"] = sun_alt_max
                    all_scheduled_obs.append(obs)

    result = np.concatenate(all_scheduled_obs)
    # Put in the scripted ID so it's easier to track which ones fail.
    result["scripted_id"] = np.arange(result.size)
    return result



__all__ = ("UnscheduledDowntimeData",)

import random

import numpy as np
from astropy.time import Time, TimeDelta


class UnscheduledDowntimeData:
    """Handle (and create) the unscheduled downtime information.

    Parameters
    ----------
    start_time : `astropy.time.Time`
        The time of the start of the simulation.
        The cloud database will be assumed to start on Jan 01 of the same year.
    seed : `int`, optional
        The random seed for creating the random nights of unscheduled downtime. Default 1516231120.
    start_of_night_offset : `float`, optional
        The fraction of a day to offset from MJD.0 to reach the defined start of a night ('noon' works).
        Default 0.16 (UTC midnight in Chile) - 0.5 (minus half a day) = -0.34
    survey_length : `int`, optional
        The number of nights in the total survey. Default 3650*2.
    """

    MINOR_EVENT = {"P": 0.0137, "length": 1, "level": "minor event"}
    INTERMEDIATE_EVENT = {"P": 0.00548, "length": 3, "level": "intermediate event"}
    MAJOR_EVENT = {"P": 0.00137, "length": 7, "level": "major event"}
    CATASTROPHIC_EVENT = {"P": 0.000274, "length": 14, "level": "catastrophic event"}

    def __init__(
        self,
        start_time,
        seed=1516231120,
        start_of_night_offset=-0.34,
        survey_length=3650 * 2,
    ):
        self.seed = seed
        self.survey_length = survey_length
        year_start = start_time.datetime.year
        self.night0 = Time("%d-01-01" % year_start, format="isot", scale="tai") + TimeDelta(
            start_of_night_offset, format="jd"
        )

        # Scheduled downtime data is a np.ndarray of start / end / activity for each scheduled downtime.
        self.downtime = None
        self.make_data()

    def __call__(self):
        """Return the array of unscheduled downtimes.

        Parameters
        ----------
        time : `astropy.time.Time`
            Time in the simulation for which to find the current downtime.

        Returns
        -------
        downtime : `np.ndarray`
            The array of all unscheduled downtimes, with keys for 'start', 'end', 'activity',
            corresponding to `astropy.time.Time`, `astropy.time.Time`, and `str`.
        """
        return self.downtime

    def _downtime_status(self, time):
        """Look behind the scenes at the downtime status/next values"""
        next_start = self.downtime["start"].searchsorted(time, side="right")
        next_end = self.downtime["end"].searchsorted(time, side="right")
        if next_start > next_end:
            current = self.downtime[next_end]
        else:
            current = None
        future = self.downtime[next_start:]
        return current, future

    def make_data(self):
        """Configure the set of unscheduled downtimes.

        This function creates the unscheduled downtimes based on a set of probabilities
        of the downtime type occurance.

        The random downtime is calculated using the following probabilities:

        minor event : remainder of night and next day = 5/365 days e.g. power supply failure
        intermediate : 3 nights = 2/365 days e.g. repair filter mechanism, rotator, hexapod, or shutter
        major event : 7 nights = 1/2*365 days
        catastrophic event : 14 nights = 1/3650 days e.g. replace a raft
        """
        random.seed(self.seed)

        starts = []
        ends = []
        acts = []
        night = 0
        while night < self.survey_length:
            prob = random.random()
            if prob < self.CATASTROPHIC_EVENT["P"]:
                start_night = self.night0 + TimeDelta(night, format="jd")
                starts.append(start_night)
                end_night = start_night + TimeDelta(self.CATASTROPHIC_EVENT["length"], format="jd")
                ends.append(end_night)
                acts.append(self.CATASTROPHIC_EVENT["level"])
                night += self.CATASTROPHIC_EVENT["length"] + 1
                continue
            else:
                prob = random.random()
                if prob < self.MAJOR_EVENT["P"]:
                    start_night = self.night0 + TimeDelta(night, format="jd")
                    starts.append(start_night)
                    end_night = start_night + TimeDelta(self.MAJOR_EVENT["length"], format="jd")
                    ends.append(end_night)
                    acts.append(self.MAJOR_EVENT["level"])
                    night += self.MAJOR_EVENT["length"] + 1
                    continue
                else:
                    prob = random.random()
                    if prob < self.INTERMEDIATE_EVENT["P"]:
                        start_night = self.night0 + TimeDelta(night, format="jd")
                        starts.append(start_night)
                        end_night = start_night + TimeDelta(self.INTERMEDIATE_EVENT["length"], format="jd")
                        ends.append(end_night)
                        acts.append(self.INTERMEDIATE_EVENT["level"])
                        night += self.INTERMEDIATE_EVENT["length"] + 1
                        continue
                    else:
                        prob = random.random()
                        if prob < self.MINOR_EVENT["P"]:
                            start_night = self.night0 + TimeDelta(night, format="jd")
                            starts.append(start_night)
                            end_night = start_night + TimeDelta(self.MINOR_EVENT["length"], format="jd")
                            ends.append(end_night)
                            acts.append(self.MINOR_EVENT["level"])
                            night += self.MINOR_EVENT["length"] + 1
            night += 1
        self.downtime = np.array(
            list(zip(starts, ends, acts)),
            dtype=[("start", "O"), ("end", "O"), ("activity", "O")],
        )

    def total_downtime(self):
        """Return total downtime (in days).

        Returns
        -------
        total : `int`
            Total number of downtime days.
        """
        total = 0
        for td in self.downtime["end"] - self.downtime["start"]:
            total += td.jd
        return total



__all__ = ("DowntimeModel",)

import warnings


class DowntimeModel:
    """Downtime estimates, both scheduled and unscheduled.

    Parameters
    ----------
    config: DowntimeModelConfig, optional
        A configuration class for the downtime model.
        This can be None, in which case the default DowntimeModelConfig is used.
        The user should set any non-default values for DowntimeModelConfig before
        configuration of the actual DowntimeModel.

    self.efd_requirements and self.target_requirements are also set.
    efd_requirements is a tuple: (list of str, float).
    This corresponds to the data columns required from the EFD and the amount of time history required.
    target_requirements is a list of str.
    This corresponds to the data columns required in the target dictionary passed when calculating the
    processed telemetry values.
    """

    def __init__(
        self,
        sched_down_col="scheduled_downtimes",
        unsched_down_col="unscheduled_downtimes",
        time_col="time",
    ):
        self.sched_down = sched_down_col
        self.unsched_down = unsched_down_col
        self.target_requirements = time_col

    def configure(self, config=None):
        warnings.warn("The configure method is deprecated.")

    def config_info(self):
        warnings.warn("The configure method is deprecated.")

    def __call__(self, efd_data, target_dict):
        """Calculate the sky coverage due to clouds.

        Parameters
        ----------
        efd_data: dict
            Dictionary of input telemetry, typically from the EFD.
            This must contain columns self.efd_requirements.
            (work in progress on handling time history).
        target_dict: dict
            Dictionary of target values over which to calculate the processed telemetry.
            (e.g. mapDict = {'ra': [], 'dec': [], 'altitude': [], 'azimuth': [], 'airmass': []})
            Here we use 'time', an astropy.time.Time, as we just need to know the time.

        Returns
        -------
        dict of bool, astropy.time.Time, astropy.time.Time
            Status of telescope (True = Down, False = Up) at time,
            time of expected end of downtime (~noon of the first available day),
            time of next scheduled downtime (~noon of the first available day).
        """
        # Check for downtime in scheduled downtimes.
        time = target_dict[self.target_requirements]
        next_start = efd_data[self.sched_down]["start"].searchsorted(time, side="right")
        next_end = efd_data[self.sched_down]["end"].searchsorted(time, side="right")
        if next_start > next_end:
            # Currently in a scheduled downtime.
            current_sched = efd_data[self.sched_down][next_end]
        else:
            # Not currently in a scheduled downtime.
            current_sched = None
        # This will be the next reported/expected downtime.
        next_sched = efd_data[self.sched_down][next_start]
        # Check for downtime in unscheduled downtimes.
        next_start = efd_data[self.unsched_down]["start"].searchsorted(time, side="right")
        next_end = efd_data[self.unsched_down]["end"].searchsorted(time, side="right")
        if next_start > next_end:
            # Currently in an unscheduled downtime.
            current_unsched = efd_data[self.unsched_down][next_end]
        else:
            current_unsched = None

        # Figure out what to report about current state.
        if current_sched is None and current_unsched is None:  # neither down
            status = False
            end_down = None
        else:  # we have a downtime from something ..
            if current_unsched is None:  # sched down only
                status = True
                end_down = current_sched["end"]
            elif current_sched is None:  # unsched down only
                status = True
                # should decide what to report on end of downtime here ..
                end_down = current_unsched["end"]
            else:  # both down ..
                status = True
                end_down = max(current_sched["end"], current_unsched["end"])
        return {"status": status, "end": end_down, "next": next_sched["start"]}



__all__ = ("generate_nights",)

import numpy as np
from astropy.coordinates import AltAz, EarthLocation, get_body, get_sun
from astropy.time import Time
from scipy.optimize import Bounds, minimize

from rubin_scheduler.utils import Site


def lin_interp(x, x0, x1, y0, y1):
    """
    Do a bunch of linear interpolations
    """
    y = y0 * (1.0 - (x - x0) / (x1 - x0)) + y1 * (x - x0) / (x1 - x0)
    return y


def alt_passing_interp(times, altitudes, goal_alt=0.0, rising=True):
    """find time when a body passses some altitude"""
    if rising:
        below = np.where((altitudes < goal_alt) & (np.roll(altitudes, -1) > goal_alt))[0]
        above = below + 1
    else:
        above = np.where((altitudes > goal_alt) & (np.roll(altitudes, -1) < goal_alt))[0]
        below = above + 1

    if (below.max() >= np.size(below)) | (above.max() >= np.size(above)):
        below = below[:-1]
        above = above[:-1]

    pass_times = lin_interp(goal_alt, altitudes[above], altitudes[below], times[above], times[below])
    return pass_times


def alt_sun_sum(in_mjds, location, offset):
    times = Time(in_mjds, format="mjd")
    sun = get_sun(times)
    aa = AltAz(location=location, obstime=times)
    sun = sun.transform_to(aa)
    result = np.sum(np.abs(sun.alt.deg + offset))
    return result


def alt_moon_sum(in_mjds, location, offset):
    times = Time(in_mjds, format="mjd")
    moon = get_body("moon", times)
    aa = AltAz(location=location, obstime=times)
    moon = moon.transform_to(aa)
    result = np.sum(np.abs(moon.alt.deg + offset))
    return result


def generate_nights(mjd_start, duration=3653.0, rough_step=2, verbose=False):
    """Generate the sunset and twilight times for a range of dates

    Parameters
    ----------
    mjd_start : float
        The starting mjd
    duration : float (3653)
        How long to compute times for (days)
    rough_step : float (2.)
        Time step for computing first pass rough sunrise times (hours)
    """

    # Let's find the nights first, find the times where the sun crosses the meridian.
    site = Site("LSST")
    location = EarthLocation(lat=site.latitude, lon=site.longitude, height=site.height)
    # go on 1/10th of a day steps
    t_step = rough_step / 24.0
    pad_around = 30.0 / 24.0
    t_sparse = Time(
        np.arange(mjd_start - pad_around, duration + mjd_start + pad_around + t_step, t_step),
        format="mjd",
        location=location,
    )
    sun = get_sun(t_sparse)
    aa = AltAz(location=location, obstime=t_sparse)
    sun_aa_sparse = sun.transform_to(aa)

    moon = get_body("moon", t_sparse)
    moon_aa_sparse = moon.transform_to(aa)

    # Indices right before sunset
    mjd_sunset_rough = alt_passing_interp(t_sparse.mjd, sun_aa_sparse.alt.deg, goal_alt=0.0, rising=False)

    names = [
        "night",
        "sunset",
        "sun_n12_setting",
        "sun_n18_setting",
        "sun_n18_rising",
        "sun_n12_rising",
        "sunrise",
        "moonrise",
        "moonset",
    ]
    types = [int]
    types.extend([float] * (len(names) - 1))
    alt_info_array = np.zeros(mjd_sunset_rough.size, dtype=list(zip(names, types)))
    alt_info_array["sunset"] = mjd_sunset_rough
    # label the nights
    alt_info_array["night"] = np.arange(mjd_sunset_rough.size)
    night_1_index = np.searchsorted(alt_info_array["sunset"], mjd_start)
    alt_info_array["night"] += 1 - alt_info_array["night"][night_1_index]

    # OK, now I have sunset times
    sunrises = alt_passing_interp(t_sparse.mjd, sun_aa_sparse.alt.deg, goal_alt=0.0, rising=True)
    # Make sure sunrise happens after sunset
    insert_indices = np.searchsorted(alt_info_array["sunset"], sunrises, side="left") - 1
    good_indices = np.where(insert_indices > 0)[0]
    alt_info_array["sunrise"][insert_indices[good_indices]] = sunrises[good_indices]

    # Should probably write the function to get rin of the copy-pasta
    point = alt_passing_interp(t_sparse.mjd, sun_aa_sparse.alt.deg, goal_alt=-12.0, rising=False)
    insert_indices = np.searchsorted(alt_info_array["sunset"], point, side="left") - 1
    good_indices = np.where(insert_indices > 0)[0]
    alt_info_array["sun_n12_setting"][insert_indices[good_indices]] = point[good_indices]

    point = alt_passing_interp(t_sparse.mjd, sun_aa_sparse.alt.deg, goal_alt=-12.0, rising=True)
    insert_indices = np.searchsorted(alt_info_array["sunset"], point, side="left") - 1
    good_indices = np.where(insert_indices > 0)[0]
    alt_info_array["sun_n12_rising"][insert_indices[good_indices]] = point[good_indices]

    point = alt_passing_interp(t_sparse.mjd, sun_aa_sparse.alt.deg, goal_alt=-18.0, rising=True)
    insert_indices = np.searchsorted(alt_info_array["sunset"], point, side="left") - 1
    good_indices = np.where(insert_indices > 0)[0]
    alt_info_array["sun_n18_rising"][insert_indices[good_indices]] = point[good_indices]

    point = alt_passing_interp(t_sparse.mjd, sun_aa_sparse.alt.deg, goal_alt=-18.0, rising=False)
    insert_indices = np.searchsorted(alt_info_array["sunset"], point, side="left") - 1
    good_indices = np.where(insert_indices > 0)[0]
    alt_info_array["sun_n18_setting"][insert_indices[good_indices]] = point[good_indices]

    point = alt_passing_interp(t_sparse.mjd, moon_aa_sparse.alt.deg, goal_alt=0.0, rising=True)
    insert_indices = np.searchsorted(alt_info_array["sunset"], point, side="left") - 1
    good_indices = np.where(insert_indices > 0)[0]
    alt_info_array["moonrise"][insert_indices[good_indices]] = point[good_indices]

    point = alt_passing_interp(t_sparse.mjd, moon_aa_sparse.alt.deg, goal_alt=0.0, rising=False)
    insert_indices = np.searchsorted(alt_info_array["sunset"], point, side="left") - 1
    good_indices = np.where(insert_indices > 0)[0]
    alt_info_array["moonset"][insert_indices[good_indices]] = point[good_indices]

    # Crop off some regions that might not have been filled
    good = np.where(alt_info_array["night"] > 0)[0]
    alt_info_array = alt_info_array[good[:-1]]

    # Put bounds so we don't go to far from rough fit
    refined_mjds = np.empty_like(alt_info_array)
    refined_mjds["night"] = alt_info_array["night"]

    names_dict = {
        "sunset": 0.0,
        "sun_n12_setting": 12.0,
        "sun_n18_setting": 18.0,
        "sun_n18_rising": 18.0,
        "sun_n12_rising": 12.0,
        "sunrise": 0,
    }

    # Need to keep the runtime reasonable
    options = {"maxiter": 5}
    for key in names_dict:
        if verbose:
            print(key)
        bounds = Bounds(
            alt_info_array[key] - rough_step / 10.0 / 24.0,
            alt_info_array[key] + rough_step / 10.0 / 24.0,
        )
        new_mjds = minimize(
            alt_sun_sum,
            alt_info_array[key],
            bounds=bounds,
            args=(location, names_dict[key]),
            options=options,
        )
        refined_mjds[key] = new_mjds.x

    for key in ["moonrise", "moonset"]:
        if verbose:
            print(key)
        bounds = Bounds(
            alt_info_array[key] - rough_step / 10.0 / 24.0,
            alt_info_array[key] + rough_step / 10.0 / 24.0,
        )
        new_mjds = minimize(
            alt_moon_sum,
            alt_info_array[key],
            bounds=bounds,
            args=(location, 0.0),
            options=options,
        )
        refined_mjds[key] = new_mjds.x

    # Note, there is the possibility that some moonrise/moonset
    # times changed nights upon refinement. I suppose
    # I could do another seatchsorted pass here just
    # to be extra sure nothing changed.

    return alt_info_array, refined_mjds


if __name__ == "__main__":
    # Let's use astropy to pre-compute the sunrise/sunset/twilight/moonrise/moonset times we're interested in.
    mjd_start = 59853.5
    #
    rough_times, refined_mjds = generate_nights(
        mjd_start - 365.25 * 2 - 40.0, duration=365.25 * 24 + 80, rough_step=2
    )
    # rough_times, refined_mjds = generate_nights(mjd_start, duration=50, rough_step=2)
    # Maybe just use pandas to dump it to a csv file?
    np.savez("night_info.npz", rough_times=rough_times, refined_mjds=refined_mjds)



__all__ = ("GreedySurvey", "BlobSurvey")

import warnings
from copy import copy

import healpy as hp
import numpy as np

from rubin_scheduler.scheduler.surveys import BaseMarkovSurvey
from rubin_scheduler.scheduler.utils import empty_observation, int_binned_stat, order_observations
from rubin_scheduler.utils import _angular_separation, _approx_ra_dec2_alt_az, _hpid2_ra_dec, hp_grow_argsort


class GreedySurvey(BaseMarkovSurvey):
    """
    Select pointings in a greedy way using a Markov Decision Process.
    """

    def __init__(
        self,
        basis_functions,
        basis_weights,
        filtername="r",
        block_size=1,
        smoothing_kernel=None,
        nside=None,
        dither=True,
        seed=42,
        ignore_obs=None,
        survey_name=None,
        nexp=2,
        exptime=30.0,
        detailers=None,
        camera="LSST",
        area_required=None,
        fields=None,
    ):
        extra_features = {}

        if survey_name == "":
            survey_name = f"Greedy {filtername}"

        super(GreedySurvey, self).__init__(
            basis_functions=basis_functions,
            basis_weights=basis_weights,
            extra_features=extra_features,
            smoothing_kernel=smoothing_kernel,
            ignore_obs=ignore_obs,
            nside=nside,
            survey_name=survey_name,
            dither=dither,
            detailers=detailers,
            camera=camera,
            area_required=area_required,
            fields=fields,
        )
        self.filtername = filtername
        self.block_size = block_size
        self.nexp = nexp
        self.exptime = exptime

    def generate_observations_rough(self, conditions):
        """
        Just point at the highest reward healpix
        """
        self.reward = self.calc_reward_function(conditions)

        # Check if we need to spin the tesselation
        if self.dither & (conditions.night != self.night):
            self._spin_fields(conditions)
            self.night = copy(conditions.night)

        # Let's find the best N from the fields
        order = np.argsort(self.reward)[::-1]
        # Crop off any NaNs
        order = order[~np.isnan(self.reward[order])]

        iter = 0
        while True:
            best_hp = order[iter * self.block_size : (iter + 1) * self.block_size]
            best_fields = np.unique(self.hp2fields[best_hp])
            observations = []
            for field in best_fields:
                obs = empty_observation()
                obs["RA"] = self.fields["RA"][field]
                obs["dec"] = self.fields["dec"][field]
                obs["rotSkyPos"] = 0.0
                obs["filter"] = self.filtername
                obs["nexp"] = self.nexp
                obs["exptime"] = self.exptime
                obs["field_id"] = -1
                obs["note"] = self.survey_name

                observations.append(obs)
                break
            iter += 1
            if len(observations) > 0 or (iter + 2) * self.block_size > len(order):
                break
        return observations


class BlobSurvey(GreedySurvey):
    """Select observations in large, mostly contiguous, blobs.

    Parameters
    ----------
    filtername1 : `str`
        The filter to observe in.
    filtername2 : `str`
        The filter to pair with the first observation. If set to None, no pair
        will be observed.
    slew_approx : `float`
        The approximate slewtime between neerby fields (seconds). Used to calculate
        how many observations can be taken in the desired time block.
    nexp : `int`
        The number of exposures to take in a visit.
    exp_dict : `dict`
        If set, should have keys of filtername and values of ints that are the nuber of exposures to take
        per visit. For estimating block time, nexp is still used.
    filter_change_approx : `float`
         The approximate time it takes to change filters (seconds).
    ideal_pair_time : `float`
        The ideal time gap wanted between observations to the same pointing (minutes)
    min_pair_time : `float`
        The minimum acceptable pair time (minutes)
    flush_time : `float`
        The time past the final expected exposure to flush the queue. Keeps observations
        from lingering past when they should be executed. (minutes)
    twilight_scale : `bool`
        Scale the block size to fill up to twilight. Set to False if running in twilight
    in_twilight : `bool`
        Scale the block size to stay within twilight time.
    check_scheduled : `bool`
        Check if there are scheduled observations and scale blob size to match
    min_area : `float`
        If set, demand the reward function have an area of so many square degrees before executing
    grow_blob : `bool`
        If True, try to grow the blob from the global maximum. Otherwise, just use a simple sort.
        Simple sort will not constrain the blob to be contiguous.
    max_radius_peak : `float`
        The maximum radius to demand things be within the maximum of the reward function. (degrees)
        Note that traveling salesman solver can have rare failures if this is set too large
        (probably issue with projection effects or something).
    """

    def __init__(
        self,
        basis_functions,
        basis_weights,
        filtername1="r",
        filtername2="g",
        slew_approx=7.5,
        filter_change_approx=140.0,
        read_approx=2.0,
        exptime=30.0,
        nexp=2,
        nexp_dict=None,
        ideal_pair_time=22.0,
        min_pair_time=15.0,
        flush_time=30.0,
        smoothing_kernel=None,
        nside=None,
        dither=True,
        seed=42,
        ignore_obs=None,
        survey_note="blob",
        detailers=None,
        camera="LSST",
        twilight_scale=True,
        in_twilight=False,
        check_scheduled=True,
        min_area=None,
        grow_blob=True,
        area_required=None,
        max_radius_peak=40.0,
        fields=None,
        survey_name="",
        **kwargs,
    ):
        self.filtername1 = filtername1
        self.filtername2 = filtername2

        super(BlobSurvey, self).__init__(
            basis_functions=basis_functions,
            basis_weights=basis_weights,
            filtername=None,
            block_size=0,
            smoothing_kernel=smoothing_kernel,
            dither=dither,
            seed=seed,
            ignore_obs=ignore_obs,
            nside=nside,
            detailers=detailers,
            camera=camera,
            area_required=area_required,
            fields=fields,
        )
        self.flush_time = flush_time / 60.0 / 24.0  # convert to days
        self.nexp = nexp
        self.nexp_dict = nexp_dict
        self.exptime = exptime
        self.slew_approx = slew_approx
        self.read_approx = read_approx
        self.hpids = np.arange(hp.nside2npix(self.nside))
        self.twilight_scale = twilight_scale
        self.in_twilight = in_twilight
        self.grow_blob = grow_blob
        self.max_radius_peak = np.radians(max_radius_peak)

        if self.twilight_scale & self.in_twilight:
            warnings.warn("Both twilight_scale and in_twilight are set to True. That is probably wrong.")

        self.min_area = min_area
        self.check_scheduled = check_scheduled
        # If we are taking pairs in same filter, no need to add filter change time.
        if filtername1 == filtername2:
            filter_change_approx = 0
        # Compute the minimum time needed to observe a blob (or observe, then repeat.)
        if filtername2 is not None:
            self.time_needed = (
                (min_pair_time * 60.0 * 2.0 + exptime + read_approx + filter_change_approx) / 24.0 / 3600.0
            )  # Days
        else:
            self.time_needed = (min_pair_time * 60.0 + exptime + read_approx) / 24.0 / 3600.0  # Days
        self.filter_set = set(filtername1)
        if filtername2 is None:
            self.filter2_set = self.filter_set
        else:
            self.filter2_set = set(filtername2)

        self.ra, self.dec = _hpid2_ra_dec(self.nside, self.hpids)

        self.survey_note = survey_note
        self.counter = 1  # start at 1, because 0 is default in empty observation
        self.min_pair_time = min_pair_time
        self.ideal_pair_time = ideal_pair_time

        self.pixarea = hp.nside2pixarea(self.nside, degrees=True)

        # If we are only using one filter, this could be useful
        if (self.filtername2 is None) | (self.filtername1 == self.filtername2):
            self.filtername = self.filtername1

    def _generate_survey_name(self):
        self.survey_name = f"Blob survey {self.filtername1}"
        if self.filtername2 is None:
            self.survey_name += f"_{self.filtername1}"
        else:
            self.survey_name += f"_{self.filtername2}"

    def _check_feasibility(self, conditions):
        """
        Check if the survey is feasable in the current conditions.
        """
        for bf in self.basis_functions:
            result = bf.check_feasibility(conditions)
            if not result:
                return result

        # If we need to check that the reward function has enough area available
        if self.min_area is not None:
            reward = 0
            for bf, weight in zip(self.basis_functions, self.basis_weights):
                basis_value = bf(conditions)
                reward += basis_value * weight
            max_reward_indx = np.min(np.where(reward == np.nanmax(reward)))
            distances = _angular_separation(
                self.ra, self.dec, self.ra[max_reward_indx], self.dec[max_reward_indx]
            )
            valid_pix = np.where((np.isnan(reward) == False) & (distances < self.max_radius_peak))[0]
            if np.size(valid_pix) * self.pixarea < self.min_area:
                result = False
        return result

    def _set_block_size(self, conditions):
        """
        Update the block size if it's getting near a break point.
        """

        # If we are trying to get things done before twilight
        if self.twilight_scale:
            available_time = conditions.sun_n18_rising - conditions.mjd
            available_time *= 24.0 * 60.0  # to minutes
            n_ideal_blocks = available_time / self.ideal_pair_time
        else:
            n_ideal_blocks = 4

        # If we are trying to get things done before a scheduled simulation
        if self.check_scheduled:
            if len(conditions.scheduled_observations) > 0:
                available_time = np.min(conditions.scheduled_observations) - conditions.mjd
                available_time *= 24.0 * 60.0  # to minutes
                n_blocks = available_time / self.ideal_pair_time
                if n_blocks < n_ideal_blocks:
                    n_ideal_blocks = n_blocks

        # If we are trying to complete before twilight ends or the night ends
        if self.in_twilight:
            at1 = conditions.sun_n12_rising - conditions.mjd
            at2 = conditions.sun_n18_setting - conditions.mjd
            times = np.array([at1, at2])
            times = times[np.where(times > 0)]
            available_time = np.min(times) if len(times) > 0 else 0.0
            available_time *= 24.0 * 60.0  # to minutes
            n_blocks = available_time / self.ideal_pair_time
            if n_blocks < n_ideal_blocks:
                n_ideal_blocks = n_blocks

        if n_ideal_blocks >= 3:
            self.nvisit_block = int(
                np.floor(
                    self.ideal_pair_time
                    * 60.0
                    / (self.slew_approx + self.exptime + self.read_approx * (self.nexp - 1))
                )
            )
        else:
            # Now we can stretch or contract the block size to allocate the
            # remainder time until twilight starts
            # We can take the remaining time and try to do 1,2, or 3 blocks.
            possible_times = available_time / np.arange(1, 4)
            diff = np.abs(self.ideal_pair_time - possible_times)
            best_block_time = np.max(possible_times[np.where(diff == np.min(diff))])
            self.nvisit_block = int(
                np.floor(
                    best_block_time
                    * 60.0
                    / (self.slew_approx + self.exptime + self.read_approx * (self.nexp - 1))
                )
            )

        # The floor can set block to zero, make it possible to to just one
        if self.nvisit_block <= 0:
            self.nvisit_block = 1

    def calc_reward_function(self, conditions):
        # Set the number of observations we are going to try and take
        self._set_block_size(conditions)
        #  Computing reward like usual with basis functions and weights
        if self._check_feasibility(conditions):
            self.reward = 0
            indx = np.arange(hp.nside2npix(self.nside))
            for bf, weight in zip(self.basis_functions, self.basis_weights):
                basis_value = bf(conditions, indx=indx)
                self.reward += basis_value * weight
            if self.smoothing_kernel is not None:
                self.smooth_reward()
        else:
            self.reward = -np.inf
            return self.reward

        if self.area_required is not None:
            max_indices = np.where(self.reward == np.nanmax(self.reward))[0]
            if np.size(max_indices) == 0:
                # This is the case if everything is masked
                self.reward = -np.inf
            else:
                max_reward_indx = np.min(max_indices)
                distances = _angular_separation(
                    self.ra,
                    self.dec,
                    self.ra[max_reward_indx],
                    self.dec[max_reward_indx],
                )
                good_area = np.where((np.abs(self.reward) >= 0) & (distances < self.max_radius_peak))[
                    0
                ].size * hp.nside2pixarea(self.nside)
                if good_area < self.area_required:
                    self.reward = -np.inf

        self.reward_checked = True
        return self.reward

    def simple_order_sort(self):
        """Fall back if we can't link contiguous blobs in the reward map"""

        # Assuming reward has already been calcualted

        potential_hp = np.where(np.isfinite(self.reward))[0]

        # Note, using nanmax, so masked pixels might be included in the pointing.
        # I guess I should document that it's not "NaN pixels can't be observed", but
        # "non-NaN pixles CAN be observed", which probably is not intuitive.
        ufields, reward_by_field = int_binned_stat(
            self.hp2fields[potential_hp], self.reward[potential_hp], statistic=np.nanmax
        )
        # chop off any nans
        not_nans = np.where(np.isfinite(reward_by_field))
        ufields = ufields[not_nans]
        reward_by_field = reward_by_field[not_nans]

        order = np.argsort(reward_by_field)[::-1]
        ufields = ufields[order]
        self.best_fields = ufields[0 : self.nvisit_block]

    def generate_observations_rough(self, conditions):
        """
        Find a good block of observations.
        """

        self.reward = self.calc_reward_function(conditions)

        # Mask off pixels that are far away from the maximum.
        max_reward_indx = np.min(np.where(self.reward == np.nanmax(self.reward)))
        distances = _angular_separation(
            self.ra, self.dec, self.ra[max_reward_indx], self.dec[max_reward_indx]
        )

        self.reward[np.where(distances > self.max_radius_peak)] = np.nan
        # Check if we need to spin the tesselation
        if self.dither & (conditions.night != self.night):
            self._spin_fields(conditions)
            self.night = copy(conditions.night)

        if self.grow_blob:
            # Note, returns highest first
            ordered_hp = hp_grow_argsort(self.reward)
            ordered_fields = self.hp2fields[ordered_hp]
            orig_order = np.arange(ordered_fields.size)
            # Remove duplicate field pointings
            _u_of, u_indx = np.unique(ordered_fields, return_index=True)
            new_order = np.argsort(orig_order[u_indx])
            best_fields = ordered_fields[u_indx[new_order]]

            if np.size(best_fields) < self.nvisit_block:
                # Let's fall back to the simple sort
                self.simple_order_sort()
            else:
                self.best_fields = best_fields[0 : self.nvisit_block]
        else:
            self.simple_order_sort()

        if len(self.best_fields) == 0:
            # everything was nans, or self.nvisit_block was zero
            return []

        # Let's find the alt, az coords of the points (right now, hopefully doesn't change much in time block)
        # Not sure why need to convert to alt,az before running TSP, but it
        # does seem to be better.
        pointing_alt, pointing_az = _approx_ra_dec2_alt_az(
            self.fields["RA"][self.best_fields],
            self.fields["dec"][self.best_fields],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
            lmst=conditions.lmst,
        )

        better_order = order_observations(pointing_az, pointing_alt)

        # XXX-TODO: Could try to roll better_order to start at the nearest/fastest slew from current position.
        observations = []
        counter2 = 0
        approx_end_time = np.size(better_order) * (
            self.slew_approx + self.exptime + self.read_approx * (self.nexp - 1)
        )
        flush_time = conditions.mjd + approx_end_time / 3600.0 / 24.0 + self.flush_time
        for i, indx in enumerate(better_order):
            field = self.best_fields[indx]
            obs = empty_observation()
            obs["RA"] = self.fields["RA"][field]
            obs["dec"] = self.fields["dec"][field]
            obs["rotSkyPos"] = 0.0
            obs["filter"] = self.filtername1
            if self.nexp_dict is None:
                obs["nexp"] = self.nexp
            else:
                obs["nexp"] = self.nexp_dict[self.filtername1]
            obs["exptime"] = self.exptime
            obs["field_id"] = -1
            obs["note"] = "%s" % (self.survey_note)
            obs["block_id"] = self.counter
            obs["flush_by_mjd"] = flush_time
            # Add the mjd for debugging
            # obs['mjd'] = conditions.mjd
            # XXX temp debugging line
            obs["survey_id"] = i
            observations.append(obs)
            counter2 += 1

        result = observations

        return result



__all__ = (
    "BaseFeature",
    "BaseSurveyFeature",
    "NObsCount",
    "NObsSurvey",
    "LastObservation",
    "LastsequenceObservation",
    "LastFilterChange",
    "NObservations",
    "CoaddedDepth",
    "LastObserved",
    "NObsNight",
    "PairInNight",
    "RotatorAngle",
    "NObservationsSeason",
    "NObsCountSeason",
    "NObservationsCurrentSeason",
    "LastNObsTimes",
    "SurveyInNight",
    "NoteInNight",
    "NoteLastObserved",
)

import healpy as hp
import numpy as np
from scipy.stats import binned_statistic

from rubin_scheduler.scheduler import utils
from rubin_scheduler.scheduler.utils import IntRounded
from rubin_scheduler.skybrightness_pre import dark_sky
from rubin_scheduler.utils import _hpid2_ra_dec, calc_season, m5_flat_sed, ra_dec2_hpid


class BaseFeature:
    """
    Base class for features.
    """

    def __init__(self, **kwargs):
        # self.feature should be a float, bool, or healpix size numpy array, or numpy masked array
        self.feature = None

    def __call__(self):
        return self.feature


class BaseSurveyFeature(BaseFeature):
    """
    Feature that tracks progreess of the survey.
    Takes observations and updates self.feature
    """

    def add_observations_array(self, observations_array, observations_hpid):
        """ """
        print(self)
        raise NotImplementedError

    def add_observation(self, observation, indx=None, **kwargs):
        """
        Parameters
        ----------
        obsevation : dict-like
            Object that contains the information about the observation
            (ra, dec, filter, mjd, etc)
        indx : `list`-like of [`int`]
            The healpixel indices that the observation overlaps.
        """
        raise NotImplementedError


class SurveyInNight(BaseSurveyFeature):
    """Keep track of how many times a survey has executed in a night."""

    def __init__(self, survey_str=""):
        self.feature = 0
        self.survey_str = survey_str
        self.night = -1

    def add_observation(self, observation, indx=None):
        if observation["night"] != self.night:
            self.night = observation["night"]
            self.feature = 0

        if self.survey_str in observation["note"]:
            self.feature += 1


class NoteInNight(BaseSurveyFeature):
    """How many times a matching note has executed in the current night."""

    def __init__(self, notes=[]):
        self.feature = 0
        self.notes = notes
        self.current_night = -100

    def add_observations_array(self, observations_array, observations_hpid):
        if self.current_night != observations_array["night"][-1]:
            self.current_night = observations_array["night"][-1].copy()
            self.feature = 0
        indx = np.where(observations_array["night"] == observations_array["night"][-1])[0]
        for ind in indx:
            if observations_array["note"][ind] in self.notes:
                self.feature += 1

    def add_observation(self, observation, indx=None):
        if self.current_night != observation["night"]:
            self.current_night = observation["night"].copy()
            self.feature = 0
        if observation["note"] in self.notes:
            self.feature += 1


class NObsCount(BaseSurveyFeature):
    """Count the number of observations. Total number, not tracked over sky

    Parameters
    ----------
    filtername : `str` (None)
        The filter to count (if None, all filters counted)
    """

    def __init__(self, filtername=None, tag=None):
        self.feature = 0
        self.filtername = filtername
        # XXX--is "tag" actually used anywhere? Maybe should remove that.
        self.tag = tag

    def add_observations_array(self, observations_array, observations_hpid):
        if self.filtername is None:
            self.feature += np.size(observations_array)
        else:
            in_filt = np.where(observations_array["filter"] == self.filtername)[0]
            self.feature += np.size(in_filt)

    def add_observation(self, observation, indx=None):
        if (self.filtername is None) and (self.tag is None):
            # Track all observations
            self.feature += 1
        elif (
            (self.filtername is not None)
            and (self.tag is None)
            and (observation["filter"][0] in self.filtername)
        ):
            # Track all observations on a specified filter
            self.feature += 1
        elif (self.filtername is None) and (self.tag is not None) and (observation["tag"][0] in self.tag):
            # Track all observations on a specified tag
            self.feature += 1
        elif (
            (self.filtername is None)
            and (self.tag is not None)
            and
            # Track all observations on a specified filter on a specified tag
            (observation["filter"][0] in self.filtername)
            and (observation["tag"][0] in self.tag)
        ):
            self.feature += 1


class NObsCountSeason(BaseSurveyFeature):
    """Count the number of observations.

    Parameters
    ----------
    filtername : `str` (None)
        The filter to count (if None, all filters counted)
    """

    def __init__(
        self,
        season,
        nside=None,
        filtername=None,
        tag=None,
        season_modulo=2,
        offset=None,
        max_season=None,
        season_length=365.25,
    ):
        self.feature = 0
        self.filtername = filtername
        self.tag = tag
        self.season = season
        self.season_modulo = season_modulo
        if offset is None:
            self.offset = np.zeros(hp.nside2npix(nside), dtype=int)
        else:
            self.offset = offset
        self.max_season = max_season
        self.season_length = season_length

    def add_observation(self, observation, indx=None):
        season = utils.season_calc(
            observation["night"],
            modulo=self.season_modulo,
            offset=self.offset[indx],
            max_season=self.max_season,
            season_length=self.season_length,
        )
        if self.season in season:
            if (self.filtername is None) and (self.tag is None):
                # Track all observations
                self.feature += 1
            elif (
                (self.filtername is not None)
                and (self.tag is None)
                and (observation["filter"][0] in self.filtername)
            ):
                # Track all observations on a specified filter
                self.feature += 1
            elif (self.filtername is None) and (self.tag is not None) and (observation["tag"][0] in self.tag):
                # Track all observations on a specified tag
                self.feature += 1
            elif (
                (self.filtername is None)
                and (self.tag is not None)
                and
                # Track all observations on a specified filter on a specified tag
                (observation["filter"][0] in self.filtername)
                and (observation["tag"][0] in self.tag)
            ):
                self.feature += 1


class NObsSurvey(BaseSurveyFeature):
    """Count the number of observations.

    Parameters
    ----------
    note : `str` (None)
        Only count observations that have str in their note field
    """

    def __init__(self, note=None):
        self.feature = 0
        self.note = note

    def add_observation(self, observation, indx=None):
        # Track all observations
        if self.note is None:
            self.feature += 1
        else:
            if self.note in observation["note"]:
                self.feature += 1


class LastObservation(BaseSurveyFeature):
    """Track the last observation. Useful if you want to see when the
    last time a survey took an observation.

    Parameters
    ----------
    survey_name : `str` (None)
        Only records if the survey name matches (or survey_name set to None)
    """

    def __init__(self, survey_name=None):
        self.survey_name = survey_name
        # Start out with an empty observation
        self.feature = utils.empty_observation()

    def add_observations_array(self, observations_array, observations_hpid):
        if self.survey_name is not None:
            good = np.where(observations_array["note"] == self.survey_name)[0]
            if np.size(good) < 0:
                self.feature = observations_array[good[-1]]
        else:
            if len(observations_array) > 0:
                self.feature = observations_array[-1]

    def add_observation(self, observation, indx=None):
        if self.survey_name is not None:
            if self.survey_name in observation["note"]:
                self.feature = observation
        else:
            self.feature = observation


class LastsequenceObservation(BaseSurveyFeature):
    """When was the last observation"""

    def __init__(self, sequence_ids=""):
        self.sequence_ids = sequence_ids  # The ids of all sequence observations...
        # Start out with an empty observation
        self.feature = utils.empty_observation()

    def add_observation(self, observation, indx=None):
        if observation["survey_id"] in self.sequence_ids:
            self.feature = observation


class LastFilterChange(BaseSurveyFeature):
    """Record when the filter last changed."""

    def __init__(self):
        self.feature = {"mjd": 0.0, "previous_filter": None, "current_filter": None}

    def add_observation(self, observation, indx=None):
        if self.feature["current_filter"] is None:
            self.feature["mjd"] = observation["mjd"][0]
            self.feature["previous_filter"] = None
            self.feature["current_filter"] = observation["filter"][0]
        elif observation["filter"][0] != self.feature["current_filter"]:
            self.feature["mjd"] = observation["mjd"][0]
            self.feature["previous_filter"] = self.feature["current_filter"]
            self.feature["current_filter"] = observation["filter"][0]


class NObservations(BaseSurveyFeature):
    """
    Track the number of observations that have been made across the sky.

    Parameters
    ----------
    filtername : `str` ('r')
        String or list that has all the filters that can count.
    nside : `int` (32)
        The nside of the healpixel map to use

    """

    def __init__(self, filtername=None, nside=None, survey_name=None):
        if nside is None:
            nside = utils.set_default_nside()

        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        self.filtername = filtername
        self.survey_name = survey_name
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.filtername is not None:
            valid_indx[np.where(observations_hpid["filter"] != self.filtername)[0]] = False
        if self.survey_name is not None:
            tmp = [name in self.survey_name for name in observations_hpid["note"]]
            valid_indx = valid_indx * np.array(tmp)
        data = observations_hpid[valid_indx]
        if np.size(data) > 0:
            result, _be, _bn = binned_statistic(
                data["hpid"], np.ones(data.size), statistic=np.sum, bins=self.bins
            )
            self.feature += result

    def add_observation(self, observation, indx=None):
        """
        Parameters
        ----------
        indx : `list`-like of [`int`]
            The indices of the healpixel map that have been observed by observation
        """

        if self.filtername is None or observation["filter"][0] in self.filtername:
            if self.survey_name is None or observation["note"] in self.survey_name:
                self.feature[indx] += 1


class NObservationsSeason(BaseSurveyFeature):
    """
    Track the number of observations that have been made across sky

    Parameters
    ----------
    season : `int`
        Only count observations in this season (year).
    filtername : `str` ('r')
        String or list that has all the filters that can count.
    nside : `int` (32)
        The nside of the healpixel map to use
    offset : `int` (0)
        The offset to use when computing the season (days)
    modulo : `int` (None)
        How to mod the years when computing season
    """

    def __init__(
        self,
        season,
        filtername=None,
        nside=None,
        offset=0,
        modulo=None,
        max_season=None,
        season_length=365.25,
    ):
        if offset is None:
            offset = np.zeros(hp.nside2npix(nside), dtype=int)
        if nside is None:
            nside = utils.set_default_nside()

        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        self.filtername = filtername
        self.offset = offset
        self.modulo = modulo
        self.season = season
        self.max_season = max_season
        self.season_length = season_length

    def add_observation(self, observation, indx=None):
        """
        Parameters
        ----------
        indx :`list`-like of [`int`]
            The indices of the healpixel map that have been observed by observation
        """

        observation_season = utils.season_calc(
            observation["night"],
            offset=self.offset[indx],
            modulo=self.modulo,
            max_season=self.max_season,
            season_length=self.season_length,
        )
        if self.season in observation_season:
            if self.filtername is None or observation["filter"][0] in self.filtername:
                self.feature[indx] += 1


class LargestN:
    def __init__(self, n):
        self.n = n

    def __call__(self, in_arr):
        if np.size(in_arr) < self.n:
            return -1
        result = in_arr[-self.n]
        return result


class LastNObsTimes(BaseSurveyFeature):
    """Record the last three observations for each healpixel"""

    def __init__(self, filtername=None, n_obs=3, nside=None):
        self.filtername = filtername
        self.n_obs = n_obs
        if nside is None:
            nside = utils.set_default_nside()
        self.feature = np.zeros((n_obs, hp.nside2npix(nside)), dtype=float)
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        # Assumes we're already sorted on mjd
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.filtername is not None:
            valid_indx[np.where(observations_hpid["filter"] != self.filtername)[0]] = False
        data = observations_hpid[valid_indx]

        if np.size(data) > 0:
            for i in range(1, self.n_obs + 1):
                func = LargestN(i)
                result, _be, _bn = binned_statistic(data["hpid"], data["mjd"], statistic=func, bins=self.bins)
                # some_vals = np.where(np.sum(result, axis=1) > 0)[0]
                self.feature[-i, :] = result

    def add_observation(self, observation, indx=None):
        if self.filtername is None or observation["filter"][0] in self.filtername:
            self.feature[0:-1, indx] = self.feature[1:, indx]
            self.feature[-1, indx] = observation["mjd"]


class NObservationsCurrentSeason(BaseSurveyFeature):
    """Track how many observations have been taken in the current season
    that meet criteria.
    """

    def __init__(
        self,
        filtername=None,
        nside=None,
        seeing_fwhm_max=None,
        m5_penalty_max=None,
        mjd_start=1,
    ):
        self.filtername = filtername
        if nside is None:
            self.nside = utils.set_default_nside()
        else:
            self.nside = nside
        self.seeing_fwhm_max = seeing_fwhm_max
        self.m5_penalty_max = m5_penalty_max

        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        if self.filtername is not None:
            self.dark_map = dark_sky(nside)[filtername]
        self.ones = np.ones(hp.nside2npix(self.nside))
        self.ra, self.dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
        self.season_map = calc_season(np.degrees(self.ra), mjd_start)
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def season_update(self, observation=None, conditions=None):
        """clear the map anywhere the season has rolled over"""
        if observation is not None:
            current_season = calc_season(np.degrees(self.ra), observation["mjd"])
        if conditions is not None:
            current_season = calc_season(np.degrees(self.ra), conditions.mjd)

        # If the season has changed anywhere, set that count to zero
        new_season = np.where((self.season_map - current_season) != 0)
        self.feature[new_season] = 0
        self.season_map = current_season

    def add_observations_array(self, observations_array, observations_hpid):
        self.season_update(observation=observations_array[-1])

        check1 = np.zeros(observations_array.size, dtype=bool)
        if self.seeing_fwhm_max is not None:
            check1[np.where(observations_array["FWHMeff"] <= self.seeing_fwhm_max)] = True
        else:
            check1[:] = True

        check2 = np.zeros(observations_array.size, dtype=bool)
        if self.m5_penalty_max is not None:
            hpid = ra_dec2_hpid(self.nside, observations_array["RA"], observations_array["dec"])
            penalty = self.dark_map[hpid] - observations_array["fivesigmadepth"]
            check2[np.where(penalty <= self.m5_penalty_max)] = True
        else:
            check2[:] = True

        if self.filtername is None:
            check3 = np.zeros(observations_array.size, dtype=bool)
            check3[np.where(observations_array["filter"] == self.filter)] = True
        else:
            check3 = np.ones(observations_array.size, dtype=bool)

        good_ids = observations_array[check1 & check2 % check3]["ID"]

        indx = np.in1d(observations_hpid["ID"], observations_array["ID"][good_ids])

        result, _be, _bn = binned_statistic(
            observations_hpid["hpid"][indx],
            observations_hpid["hpid"][indx],
            bins=self.bins,
            statistic=np.size,
        )
        self.feature += result

    def add_observation(self, observation, indx=None):
        self.season_update(observation=observation)

        if self.seeing_fwhm_max is not None:
            check1 = observation["FWHMeff"] <= self.seeing_fwhm_max
        else:
            check1 = True

        if self.m5_penalty_max is not None:
            hpid = ra_dec2_hpid(self.nside, observation["RA"], observation["dec"])
            penalty = self.dark_map[hpid] - observation["fivesigmadepth"]
            check2 = penalty <= self.m5_penalty_max
        else:
            check2 = True

        if check1 & check2:
            if self.filtername is None or observation["filter"][0] in self.filtername:
                self.feature[indx] += 1


class CoaddedDepth(BaseSurveyFeature):
    """Track the co-added depth that has been reached across the sky

    Parameters
    ----------
    fwh_meff_limit : `float` (100)
        The effective FWHM of the seeing (arcsecond).
        Images will only be added to the coadded depth if the observation FWHM
        is less than or equal to the limit.  Default 100.
    """

    def __init__(self, filtername="r", nside=None, fwh_meff_limit=100.0):
        if nside is None:
            nside = utils.set_default_nside()
        self.filtername = filtername
        self.fwh_meff_limit = IntRounded(fwh_meff_limit)
        # Starting at limiting mag of zero should be fine.
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)

    def add_observation(self, observation, indx=None):
        if observation["filter"] == self.filtername:
            if IntRounded(observation["FWHMeff"]) <= self.fwh_meff_limit:
                m5 = m5_flat_sed(
                    observation["filter"],
                    observation["skybrightness"],
                    observation["FWHMeff"],
                    observation["exptime"],
                    observation["airmass"],
                )

                self.feature[indx] = 1.25 * np.log10(10.0 ** (0.8 * self.feature[indx]) + 10.0 ** (0.8 * m5))


class LastObserved(BaseSurveyFeature):
    """
    Track when a pixel was last observed.
    Assumes observations are added in chronological order.
    """

    def __init__(self, filtername="r", nside=None, fill=np.nan):
        if nside is None:
            nside = utils.set_default_nside()

        self.filtername = filtername
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float) + fill
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        # Assumes we're already sorted on mjd
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.filtername is not None:
            valid_indx[np.where(observations_hpid["filter"] != self.filtername)[0]] = False
        data = observations_hpid[valid_indx]

        if np.size(data) > 0:
            result, _be, _bn = binned_statistic(data["hpid"], data["mjd"], statistic=np.max, bins=self.bins)
            good = np.where(result > 0)
            self.feature[good] = result[good]

    def add_observation(self, observation, indx=None):
        if self.filtername is None:
            self.feature[indx] = observation["mjd"]
        elif observation["filter"][0] in self.filtername:
            self.feature[indx] = observation["mjd"]


class NoteLastObserved(BaseSurveyFeature):
    """Track the last time an observation with a particular `note` field was
    made.

    Parameters
    ----------
    note : `str`
        Substring to match an observation note field to keep track of.
    """

    def __init__(self, note, filtername=None):
        self.note = note
        self.filtername = filtername
        self.feature = None

    def add_observation(self, observation, indx=None):
        if self.note in observation["note"] and (
            self.filtername is None or self.filtername == observation["filter"]
        ):
            self.feature = observation["mjd"]


class NObsNight(BaseSurveyFeature):
    """
    Track how many times something has been observed in a night
    (Note, even if there are two, it might not be a good pair.)

    Parameters
    ----------
    filtername : `str` ('r')
        Filter to track.
    nside : `int` (32)
        Scale of the healpix map

    """

    def __init__(self, filtername="r", nside=None):
        if nside is None:
            nside = utils.set_default_nside()

        self.filtername = filtername
        self.feature = np.zeros(hp.nside2npix(nside), dtype=int)
        self.night = None

    def add_observation(self, observation, indx=None):
        if observation["night"] != self.night:
            self.feature *= 0
            self.night = observation["night"]
        if (self.filtername == "") | (self.filtername is None):
            self.feature[indx] += 1
        elif observation["filter"][0] in self.filtername:
            self.feature[indx] += 1


class PairInNight(BaseSurveyFeature):
    """
    Track how many pairs have been observed within a night

    Parameters
    ----------
    gap_min : `float` (25.)
        The minimum time gap to consider a successful pair in minutes
    gap_max : `float` (45.)
        The maximum time gap to consider a successful pair (minutes)
    """

    def __init__(self, filtername="r", nside=None, gap_min=25.0, gap_max=45.0):
        if nside is None:
            nside = utils.set_default_nside()

        self.filtername = filtername
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        self.indx = np.arange(self.feature.size)
        self.last_observed = LastObserved(filtername=filtername)
        self.gap_min = IntRounded(gap_min / (24.0 * 60))  # Days
        self.gap_max = IntRounded(gap_max / (24.0 * 60))  # Days
        self.night = 0
        # Need to keep a full record of times and healpixels observed in a night.
        self.mjd_log = []
        self.hpid_log = []

    def add_observations_array(self, observations_array, observations_hpid):
        # ok, let's just find the largest night and toss all those in one at a time
        most_recent_night = np.where(observations_hpid["night"] == np.max(observations_hpid["night"]))[0]
        obs_hpid = observations_hpid[most_recent_night]
        uid = np.unique(obs_hpid["ID"])
        for ind_id in uid:
            # maybe a faster searchsorted way to do this, but it'll work for now
            good = np.where(obs_hpid["ID"] == ind_id)[0]
            self.add_observation(observations_hpid[good][0], observations_hpid[good]["hpid"])

    def add_observation(self, observation, indx=None):
        if self.filtername is None:
            infilt = True
        else:
            infilt = observation["filter"][0] in self.filtername
        if infilt:
            if indx is None:
                indx = self.indx
            # Clear values if on a new night
            if self.night != observation["night"]:
                self.feature *= 0.0
                self.night = observation["night"]
                self.mjd_log = []
                self.hpid_log = []

            # Look for the mjds that could possibly pair with observation
            mjd_diff = IntRounded(observation["mjd"] - np.array(self.mjd_log))
            # normally would use np.searchsorted, but need to use IntRounded
            # to be sure we are cross-platform repeatable.
            in_range_indx = np.where((mjd_diff > self.gap_min) & (mjd_diff < self.gap_max))[0]
            # Now check if any of the healpixels taken in the time gap
            # match the healpixels of the observation.
            if in_range_indx.size > 0:
                left = in_range_indx.min()
                right = in_range_indx.max() + 1
                matches = np.in1d(indx, self.hpid_log[left:right])
                self.feature[np.array(indx)[matches]] += 1

            # record the mjds and healpixels that were observed
            self.mjd_log.extend([np.max(observation["mjd"])] * np.size(indx))
            self.hpid_log.extend(list(indx))


class RotatorAngle(BaseSurveyFeature):
    """
    Track what rotation angles things are observed with.
    XXX-under construction
    """

    def __init__(self, filtername="r", binsize=10.0, nside=None):
        """"""
        if nside is None:
            nside = utils.set_default_nside()

        self.filtername = filtername
        # Actually keep a histogram at each healpixel
        self.feature = np.zeros((hp.nside2npix(nside), 360.0 / binsize), dtype=float)
        self.bins = np.arange(0, 360 + binsize, binsize)

    def add_observation(self, observation, indx=None):
        if observation["filter"][0] == self.filtername:
            # I think this is how to broadcast things properly.
            self.feature[indx, :] += np.histogram(observation.rotSkyPos, bins=self.bins)[0]



import unittest

import numpy as np

import rubin_scheduler.utils as utils


class KdTreeTestCase(unittest.TestCase):
    def test_kd_tree_api(self):
        """
        Make sure the API provided by scipy to the kdTree algorithm is functional.
        """
        _ra = np.linspace(0.0, 2.0 * np.pi)
        _dec = np.linspace(-np.pi, np.pi)

        ra, dec = np.meshgrid(_ra, _dec)
        tree = utils._build_tree(ra.flatten(), dec.flatten())

        x, y, z = utils._xyz_from_ra_dec(_ra, _dec)
        indx = tree.query_ball_point(list(zip(x, y, z)), utils.xyz_angular_radius())

        self.assertEqual(indx.shape, _ra.shape)


if __name__ == "__main__":
    unittest.main()



__all__ = ("FilterSwapScheduler", "SimpleFilterSched", "FilterSchedUzy")

from rubin_scheduler.scheduler.utils import IntRounded


class FilterSwapScheduler:
    """A simple way to schedule what filter to load"""

    def __init__(self):
        pass

    def add_observation(self, observation):
        pass

    def __call__(self, conditions):
        """
        Returns
        -------
        list of strings for the filters that should be loaded
        """
        pass


class SimpleFilterSched(FilterSwapScheduler):
    def __init__(self, illum_limit=10.0):
        self.illum_limit_ir = IntRounded(illum_limit)

    def __call__(self, conditions):
        if IntRounded(conditions.moon_phase) > self.illum_limit_ir:
            result = ["g", "r", "i", "z", "y"]
        else:
            result = ["u", "g", "r", "i", "z"]
        return result


class FilterSchedUzy(FilterSwapScheduler):
    """
    remove u in bright time. Alternate between removing z and y in dark time.

    Note, this might not work properly if we need to restart a bunch. So a more robust
    way of scheduling filter loading might be in order.
    """

    def __init__(self, illum_limit=10.0):
        self.illum_limit_ir = IntRounded(illum_limit)
        self.last_swap = 0

        self.bright_time = ["g", "r", "i", "z", "y"]
        self.dark_times = [["u", "g", "r", "i", "y"], ["u", "g", "r", "i", "z"]]

    def __call__(self, conditions):
        if IntRounded(conditions.moon_phase) > self.illum_limit_ir:
            result = self.bright_time
        else:
            indx = self.last_swap % 2
            result = self.dark_times[indx]
            if result != conditions.mounted_filters:
                self.last_swap += 1
        return result



__all__ = (
    "SolarElongMaskBasisFunction",
    "ZenithMaskBasisFunction",
    "ZenithShadowMaskBasisFunction",
    "HaMaskBasisFunction",
    "MoonAvoidanceBasisFunction",
    "MapCloudBasisFunction",
    "PlanetMaskBasisFunction",
    "MaskAzimuthBasisFunction",
    "SolarElongationMaskBasisFunction",
    "AreaCheckMaskBasisFunction",
    "AltAzShadowMaskBasisFunction",
)

import warnings

import healpy as hp
import numpy as np

from rubin_scheduler.scheduler.basis_functions import BaseBasisFunction
from rubin_scheduler.scheduler.utils import HpInLsstFov, IntRounded
from rubin_scheduler.utils import Site, _angular_separation, _hpid2_ra_dec


class SolarElongMaskBasisFunction(BaseBasisFunction):
    """Mask regions larger than some solar elongation limit

    Parameters
    ----------
    elong_limit : float (45)
        The limit beyond which to mask (degrees)
    """

    def __init__(self, elong_limit=45.0, nside=32):
        super(SolarElongMaskBasisFunction, self).__init__(nside=nside)
        self.elong_limit = IntRounded(np.radians(elong_limit))
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        to_mask = np.where(IntRounded(conditions.solar_elongation) > self.elong_limit)[0]
        result[to_mask] = np.nan
        return result


class HaMaskBasisFunction(BaseBasisFunction):
    """Limit the sky based on hour angle

    Parameters
    ----------
    ha_min : float (None)
        The minimum hour angle to accept (hours)
    ha_max : float (None)
        The maximum hour angle to accept (hours)
    """

    def __init__(self, ha_min=None, ha_max=None, nside=32):
        super(HaMaskBasisFunction, self).__init__(nside=nside)
        self.ha_max = ha_max
        self.ha_min = ha_min
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, **kwargs):
        result = self.result.copy()

        if self.ha_min is not None:
            good = np.where(conditions.HA < (self.ha_min / 12.0 * np.pi))[0]
            result[good] = np.nan
        if self.ha_max is not None:
            good = np.where(conditions.HA > (self.ha_max / 12.0 * np.pi))[0]
            result[good] = np.nan

        return result


class AreaCheckMaskBasisFunction(BaseBasisFunction):
    """Take a list of other mask basis functions, and do an additional check for area available"""

    def __init__(self, bf_list, nside=32, min_area=1000.0):
        super(AreaCheckMaskBasisFunction, self).__init__(nside=nside)
        self.bf_list = bf_list
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.min_area = min_area

    def check_feasibility(self, conditions):
        result = True
        for bf in self.bf_list:
            if not bf.check_feasibility(conditions):
                return False

        area_map = self.result.copy()
        for bf in self.bf_list:
            area_map *= bf(conditions)

        good_pix = np.where(area_map == 0)[0]
        if hp.nside2pixarea(self.nside, degrees=True) * good_pix.size < self.min_area:
            result = False
        return result

    def _calc_value(self, conditions, **kwargs):
        result = self.result.copy()
        for bf in self.bf_list:
            result *= bf(conditions)
        return result


class SolarElongationMaskBasisFunction(BaseBasisFunction):
    """Mask things at various solar elongations

    Parameters
    ----------
    min_elong : float (0)
        The minimum solar elongation to consider (degrees).
    max_elong : float (60.)
        The maximum solar elongation to consider (degrees).
    """

    def __init__(self, min_elong=0.0, max_elong=60.0, nside=None, penalty=np.nan):
        super(SolarElongationMaskBasisFunction, self).__init__(nside=nside)
        self.min_elong = np.radians(min_elong)
        self.max_elong = np.radians(max_elong)
        self.penalty = penalty
        self.result = np.empty(hp.nside2npix(self.nside), dtype=float)
        self.result.fill(self.penalty)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        in_range = np.where(
            (IntRounded(conditions.solar_elongation) >= IntRounded(self.min_elong))
            & (IntRounded(conditions.solar_elongation) <= IntRounded(self.max_elong))
        )[0]
        result[in_range] = 1
        return result


class ZenithMaskBasisFunction(BaseBasisFunction):
    """Just remove the area near zenith.

    Parameters
    ----------
    min_alt : float (20.)
        The minimum possible altitude (degrees)
    max_alt : float (82.)
        The maximum allowed altitude (degrees)
    """

    def __init__(self, min_alt=20.0, max_alt=82.0, nside=None):
        super(ZenithMaskBasisFunction, self).__init__(nside=nside)
        self.update_on_newobs = False
        self.min_alt = np.radians(min_alt)
        self.max_alt = np.radians(max_alt)
        self.result = np.empty(hp.nside2npix(self.nside), dtype=float).fill(self.penalty)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        alt_limit = np.where(
            (IntRounded(conditions.alt) > IntRounded(self.min_alt))
            & (IntRounded(conditions.alt) < IntRounded(self.max_alt))
        )[0]
        result[alt_limit] = 1
        return result


class PlanetMaskBasisFunction(BaseBasisFunction):
    """Mask the bright planets

    Parameters
    ----------
    mask_radius : float (3.5)
        The radius to mask around a planet (degrees).
    planets : list of str (None)
        A list of planet names to mask. Defaults to ['venus', 'mars', 'jupiter']. Not including
        Saturn because it moves really slow and has average apparent mag of ~0.4, so fainter than Vega.

    """

    def __init__(self, mask_radius=3.5, planets=None, nside=None, scale=1e5):
        super(PlanetMaskBasisFunction, self).__init__(nside=nside)
        if planets is None:
            planets = ["venus", "mars", "jupiter"]
        self.planets = planets
        self.mask_radius = np.radians(mask_radius)
        self.result = np.zeros(hp.nside2npix(nside))
        # set up a kdtree. Could maybe use healpy.query_disc instead.
        self.in_fov = HpInLsstFov(nside=nside, fov_radius=mask_radius, scale=scale)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        for pn in self.planets:
            indices = self.in_fov(
                np.max(conditions.planet_positions[pn + "_RA"]),
                np.max(conditions.planet_positions[pn + "_dec"]),
            )
            result[indices] = np.nan

        return result


class AltAzShadowMaskBasisFunction(BaseBasisFunction):
    """Mask out range altitudes and azimuths, then extend the
    mask so if observations are taken in pairs, the second in the pair will
    not have moved into a masked region.

    Masks any alt/az regions as specified by the conditions object, then
    applies any additional altitude masking as suppied by the kwargs.
    This mask is then extended using `shadow minutes`.

    Parameters
    ----------
    nside : `int`
        HEALpix nside. Default None will look up the package-wide default.
    min_alt : `float`
        Minimum altitude to apply to the mask. Default 20 (degrees).
    max_alt : `float`
        Maximum altitude to allow. Default 82 (degrees).
    shadow_minutes : `float`
        How long to extend masked area in longitude. Default 40 (minutes).
    """

    def __init__(
        self,
        nside=None,
        min_alt=20.0,
        max_alt=82.0,
        shadow_minutes=40.0,
    ):
        super(AltAzShadowMaskBasisFunction, self).__init__(nside=nside)
        self.min_alt = np.radians(min_alt)
        self.max_alt = np.radians(max_alt)
        self.shadow_time = shadow_minutes / 60.0 / 24.0  # To days

    def _calc_value(self, conditions, indx=None):
        # Mask everything to start
        result = np.zeros(hp.nside2npix(self.nside), dtype=float) + np.nan

        in_range_alt = np.zeros(hp.nside2npix(self.nside), dtype=int)
        in_range_az = np.zeros(hp.nside2npix(self.nside), dtype=int)

        # Compute the alt,az values in the future. Use the conditions object
        # so the results are cached and can be used by other surveys is needed.
        # Technically this could fail if the masked region is very narrow or shadow time
        # is very large.
        future_alt, future_az = conditions.future_alt_az(np.max(conditions.mjd + self.shadow_time))

        # apply limits from the conditions object
        for limits in conditions.tel_alt_limits:
            good = np.where(
                (IntRounded(conditions.alt) >= IntRounded(np.min(limits)))
                & (IntRounded(conditions.alt) <= IntRounded(np.max(limits)))
            )[0]
            in_range_alt[good] += 1
            good = np.where(
                (IntRounded(future_alt) >= IntRounded(np.min(limits)))
                & (IntRounded(future_alt) <= IntRounded(np.max(limits)))
            )[0]
            in_range_alt[good] += 1

        for limits in conditions.tel_az_limits:
            good = np.where(
                (IntRounded(conditions.az) >= IntRounded(np.min(limits)))
                & (IntRounded(conditions.az) <= IntRounded(np.max(limits)))
            )[0]
            in_range_az[good] += 1
            good = np.where(
                (IntRounded(future_az) >= IntRounded(np.min(limits)))
                & (IntRounded(future_az) <= IntRounded(np.max(limits)))
            )[0]
            in_range_az[good] += 1

        passed_all = np.where((in_range_alt > 1) & (in_range_az > 1))[0]
        result[passed_all] = 0

        # Apply additional alt constraint in case we want to be more conservative than the limit
        result[np.where(IntRounded(conditions.alt) < IntRounded(self.min_alt))] = np.nan
        result[np.where(IntRounded(conditions.alt) > IntRounded(self.max_alt))] = np.nan

        result[np.where(IntRounded(future_alt) < IntRounded(self.min_alt))] = np.nan
        result[np.where(IntRounded(future_alt) > IntRounded(self.max_alt))] = np.nan

        return result


class ZenithShadowMaskBasisFunction(BaseBasisFunction):
    """Mask the zenith, and things that will soon pass near zenith. Useful for making sure
    observations will not be too close to zenith when they need to be observed again (e.g. for a pair).

    Parameters
    ----------
    min_alt : float (20.)
        The minimum alititude to alow. Everything lower is masked. (degrees)
    max_alt : float (82.)
        The maximum altitude to alow. Everything higher is masked. (degrees)
    shadow_minutes : float (40.)
        Mask anything that will pass through the max alt in the next shadow_minutes time. (minutes)
    """

    def __init__(
        self,
        nside=None,
        min_alt=20.0,
        max_alt=82.0,
        shadow_minutes=40.0,
        penalty=np.nan,
        site="LSST",
    ):
        warnings.warn(
            "Deprecating ZenithShadowMaskBasisFunction in favor of AltAzShadowMaskBasisFunction.",
            DeprecationWarning,
        )

        super(ZenithShadowMaskBasisFunction, self).__init__(nside=nside)
        self.update_on_newobs = False

        self.penalty = penalty

        self.min_alt = np.radians(min_alt)
        self.max_alt = np.radians(max_alt)
        self.ra, self.dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
        self.shadow_minutes = np.radians(shadow_minutes / 60.0 * 360.0 / 24.0)
        # Compute the declination band where things could drift into zenith
        self.decband = np.zeros(self.dec.size, dtype=float)
        self.zenith_radius = np.radians(90.0 - max_alt) / 2.0
        site = Site(name=site)
        self.lat_rad = site.latitude_rad
        self.lon_rad = site.longitude_rad
        self.decband[
            np.where(
                (IntRounded(self.dec) < IntRounded(self.lat_rad + self.zenith_radius))
                & (IntRounded(self.dec) > IntRounded(self.lat_rad - self.zenith_radius))
            )
        ] = 1

        self.result = np.empty(hp.nside2npix(self.nside), dtype=float)
        self.result.fill(self.penalty)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        alt_limit = np.where(
            (IntRounded(conditions.alt) > IntRounded(self.min_alt))
            & (IntRounded(conditions.alt) < IntRounded(self.max_alt))
        )[0]
        result[alt_limit] = 1
        to_mask = np.where(
            (IntRounded(conditions.HA) > IntRounded(2.0 * np.pi - self.shadow_minutes - self.zenith_radius))
            & (self.decband == 1)
        )
        result[to_mask] = np.nan
        return result


class MoonAvoidanceBasisFunction(BaseBasisFunction):
    """Avoid looking too close to the moon.

    Parameters
    ----------
    moon_distance: float (30.)
        Minimum allowed moon distance. (degrees)

    XXX--TODO:  This could be a more complicated function of filter and moon phase.
    """

    def __init__(self, nside=None, moon_distance=30.0):
        super(MoonAvoidanceBasisFunction, self).__init__(nside=nside)
        self.update_on_newobs = False

        self.moon_distance = IntRounded(np.radians(moon_distance))
        self.result = np.ones(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()

        angular_distance = _angular_separation(
            conditions.az, conditions.alt, conditions.moon_az, conditions.moon_alt
        )

        result[IntRounded(angular_distance) < self.moon_distance] = np.nan

        return result


class BulkCloudBasisFunction(BaseBasisFunction):
    """Mark healpixels on a map if their cloud values are greater than
    the same healpixels on a maximum cloud map.

    Parameters
    ----------
    nside: int (default_nside)
        The healpix resolution.
    max_cloud_map : numpy array (None)
        A healpix map showing the maximum allowed cloud values for all points on the sky
    out_of_bounds_val : float (10.)
        Point value to give regions where there are no observations requested
    """

    def __init__(self, nside=None, max_cloud_map=None, max_val=0.7, out_of_bounds_val=np.nan):
        super(BulkCloudBasisFunction, self).__init__(nside=nside)
        self.update_on_newobs = False

        if max_cloud_map is None:
            self.max_cloud_map = np.zeros(hp.nside2npix(nside), dtype=float) + max_val
        else:
            self.max_cloud_map = max_cloud_map
        self.out_of_bounds_area = np.where(self.max_cloud_map > 1.0)[0]
        self.out_of_bounds_val = out_of_bounds_val
        self.result = np.ones(hp.nside2npix(self.nside))

    def _calc_value(self, conditions, indx=None):
        """
        Parameters
        ----------
        indx : list (None)
            Index values to compute, if None, full map is computed
        Returns
        -------
        Healpix map where pixels with a cloud value greater than the max_cloud_map
        value are marked as unseen.
        """

        result = self.result.copy()

        clouded = np.where(self.max_cloud_map <= conditions.bulk_cloud)
        result[clouded] = self.out_of_bounds_val

        return result


class MapCloudBasisFunction(BaseBasisFunction):
    """Mark healpixels on a map if their cloud values are greater than
    the same healpixels on a maximum cloud map. Currently a placeholder for
    when the telemetry stream can include a full sky cloud map.

    Parameters
    ----------
    nside: int (default_nside)
        The healpix resolution.
    max_cloud_map : numpy array (None)
        A healpix map showing the maximum allowed cloud values for all points on the sky
    out_of_bounds_val : float (10.)
        Point value to give regions where there are no observations requested
    """

    def __init__(self, nside=None, max_cloud_map=None, max_val=0.7, out_of_bounds_val=np.nan):
        super(BulkCloudBasisFunction, self).__init__(nside=nside)
        self.update_on_newobs = False

        if max_cloud_map is None:
            self.max_cloud_map = np.zeros(hp.nside2npix(nside), dtype=float) + max_val
        else:
            self.max_cloud_map = max_cloud_map
        self.out_of_bounds_area = np.where(self.max_cloud_map > 1.0)[0]
        self.out_of_bounds_val = out_of_bounds_val
        self.result = np.ones(hp.nside2npix(self.nside))

    def _calc_value(self, conditions, indx=None):
        """
        Parameters
        ----------
        indx : list (None)
            Index values to compute, if None, full map is computed
        Returns
        -------
        Healpix map where pixels with a cloud value greater than the max_cloud_map
        value are marked as unseen.
        """

        result = self.result.copy()

        clouded = np.where(self.max_cloud_map <= conditions.bulk_cloud)
        result[clouded] = self.out_of_bounds_val

        return result


class MaskAzimuthBasisFunction(BaseBasisFunction):
    """Mask pixels based on azimuth"""

    def __init__(self, nside=None, out_of_bounds_val=np.nan, az_min=0.0, az_max=180.0):
        super(MaskAzimuthBasisFunction, self).__init__(nside=nside)
        self.az_min = IntRounded(np.radians(az_min))
        self.az_max = IntRounded(np.radians(az_max))
        self.out_of_bounds_val = out_of_bounds_val
        self.result = np.ones(hp.nside2npix(self.nside))

    def _calc_value(self, conditions, indx=None):
        to_mask = np.where(
            (IntRounded(conditions.az) > self.az_min) & (IntRounded(conditions.az) < self.az_max)
        )[0]
        result = self.result.copy()
        result[to_mask] = self.out_of_bounds_val

        return result
